# -*- coding: utf-8 -*-

"""A dictionary-like object for holding bonds

Based on tables in an SQLite database.
"""

import logging
import sqlite3
from typing import Any, Dict, TypeVar

import numpy as np
import pandas

from molsystem.table import _Table as Table
from molsystem.frozencolumn import _FrozenColumn as FrozenColumn
from molsystem.column import _Column as Column

System_tp = TypeVar("System_tp", "System", "Bonds", None)

logger = logging.getLogger(__name__)


class _Bonds(Table):
    """The Bonds class describes the bonds in the system.

    Three attributes are required: the atoms 'i' and 'j' of the bond, and the
    bond order, which defaults to a single bond.

    Since there are two possible representations for a bond, i-j and j-i,
    the bonds are stored with i < j in order to make the indexing unique.

    Other attributes can be created, either from a list of predefined
    ones or by specifying the metadata required of an attribute. Attributes can
    also be removed. See the method 'add_attribute' for more detail.

    Bonds can be added ('append') or removed ('delete').

    In order to handle changes in bonding (and numbers of atoms) the
    system class has several tables covering the atoms and bonds. See
    the main documentation for a more detailed description. The key
    tables are:

    configuration -- a list of configurations in e.g. a trajectory
    configuration_subset -- connnects configurations with subset

    template -- a simple list of templates
    templateatom -- holds the atoms in a template, if present.
    templatebond -- holds the bonds between the template atoms, if present

    subset -- an instantiation of a template, connected with one or more
              configurations of the system.
    subset_atom -- connects the subset to the atoms, and optionally connects
                   the atom to a template atom.

    atom -- the nonvarying part of the description of the atoms
    coordinates -- the varying part of the description of atoms

    This class handles the bonds for the main system, but it does so through
    subset "all", which is a special subset that contains all of the atoms in
    the system. This subset is defined by one or more templates "all" which are
    instantiated in the subset table and connected to configurations via the
    configuration_subset table.

    Underneath, this is handled by a subset "all", which is a special
    subset that contains all of the atoms in the system. This subset is
    defined by one or more templates "all" which are instantiated in the
    subset table and connected to configurations via the
    configuration_subset table.

    If the number of atoms in the system is fixed and there are no bonds
    or they don't change (the case in classical MD), there will be one
    template "all" instantiated as one subset. All configurations will
    point to this one subset. If there are no bonds in the system, then
    this is all that is needed. However if there are bonds, the atoms
    involved in bonds need to be in the templateatom table and the bonds
    entered in the templatebond table.

    If the bonds change over time, then for each bond configuration
    there needs to be a distinct "all" template, with the appropriate
    templateatoms and templatebonds. This template is instantiated as a
    new subset and connected to one or more configurations.

    If the number of atoms changes over time, each new set of atoms also
    requires a new template and subset connected to the atoms in that
    configuration. A subset may persist over a number of configurations,
    but each time the atoms change a new subset is required.

    If both the atoms and the bonds change, then a corresponding subset
    is needed each time either changes.
    """

    def __init__(self, system: System_tp, table: str = 'templatebond') -> None:

        super().__init__(system, table)

        self._bond_db = None

    def __getitem__(self, key):
        return self.get_column(key)

    def __eq__(self, other) -> Any:
        """Return a boolean if this object is equal to another"""
        raise NotImplementedError()

    @property
    def bond_db(self):
        if self._bond_db is None:
            self._bond_db = sqlite3.connect(self._system._filename)
            self._bond_db.row_factory = self._row_factory
            self._bond_db.execute('PRAGMA foreign_keys = ON')
        return self._bond_db

    def append(
        self,
        configuration: int = None,
        bonds=None,
        **kwargs: Dict[str, Any]
    ) -> None:
        """Append one or more bonds

        The keys give the field for the data. If an existing field is not
        mentioned, then the default value is used, unless the default is None,
        in which case an error is thrown. It is an error if there is not a
        field corrresponding to a key.

        This is where it becomes a bit complicated. The atoms i & j are atoms
        in the main 'atom' table; however, the bonds are stored in
        'templatebond' referencing atoms in 'templateatom' for the template
        'all'. The real atoms are connected to the template atoms via the
        'subset_atom' table, which has an extra field referencing the template
        atom.
        """

        if bonds is not None:
            if len(kwargs) > 0:
                raise RuntimeError(
                    'Please do not give both bonds and arrays of data!'
                )
            if isinstance(bonds, sqlite3.Row):
                # One bond
                kwargs = {}
                for key, value in zip(bonds.keys(), bonds):
                    kwargs[key] = value
            else:
                first = True
                for bond in bonds:
                    if first:
                        keys = bond.keys()
                        for key, value in zip(keys, bond):
                            kwargs[key] = [value]
                        first = False
                    else:
                        for key, value in zip(keys, bond):
                            kwargs[key].append(value)

        # Check keys and lengths of added bonds
        if 'i' not in kwargs or 'j' not in kwargs:
            raise KeyError("The atoms i & j are required!")

        i = kwargs.pop('i')
        j = kwargs.pop('j')
        try:
            len_i = len(i)
        except TypeError:
            len_i = 1
            i = (i,)
        try:
            len_j = len(j)
        except TypeError:
            len_j = 1
            j = (j,)

        if len_i == 0 and len_j == 0:
            return

        if len_i == 1:
            if len_j > 1:
                i = len_j * [i[0]]
        elif len_j == 1:
            j = len_i * [j[0]]
        elif len_i != len_j:
            raise IndexError(
                f'key "j" has the wrong number of values, {len_j}. '
                f'Should be 1 or the number of values in i, {len_i}.'
            )
        # Ensure that i < j

        if isinstance(i, np.ndarray):
            i = i.tolist()
        if isinstance(j, np.ndarray):
            j = j.tolist()

        i2 = []
        j2 = []
        for i_, j_ in zip(i, j):
            # will need to handle offsets here at some point
            if not isinstance(i_, int) or not isinstance(j_, int):
                raise TypeError(
                    "'i' and 'j', the atom indices, must be integers"
                )
            if i_ < j_:
                i2.append(i_)
                j2.append(j_)
            else:
                i2.append(j_)
                j2.append(i_)

        # Get the template atoms corresponding to the i and j atoms,
        # adding them if needed.
        if configuration is None:
            configuration = self._system.current_configuration
        subset = self._system.all_subset(configuration)
        template = self._system.all_template(configuration)

        # First get current map of atoms to templateatoms
        map = {}
        for row in self.db.execute(
            "SELECT atom, templateatom FROM subset_atom"
            " WHERE subset = ?", (subset,)
        ):
            templateatom = row['templateatom']
            if templateatom is not None:
                map[row['atom']] = templateatom

        # Use dictionary because there may be duplicates
        missing = {}
        for i_ in i2:
            if i_ not in map:
                missing[i_] = None
        for j_ in j2:
            if j_ not in map:
                missing[j_] = None

        if len(missing) > 0:
            # Need to add to the all template.
            table = self._system['templateatom']
            templateatoms = table.append(n=len(missing), template=template)

            parameters = []
            # The sorting here is to keep the atoms in the same order,
            # which is what the user expects.
            for i_, tatom in zip(sorted(missing), templateatoms):
                map[i_] = tatom
                parameters.append((tatom, i_, subset))
            self.cursor.executemany(
                "UPDATE subset_atom SET templateatom = ?"
                " WHERE atom = ? AND subset = ?", parameters
            )

        # get the lists of template atoms for the bonds
        ti = []
        tj = []
        for i_ in i2:
            ti.append(map[i_])
        for j_ in j2:
            tj.append(map[j_])

        # and ... finally ... add the bonds
        if 'bondorder' in kwargs:
            bondorders = kwargs.pop('bondorder')
            self._system['templatebond'].append(
                i=ti, j=tj, bondorder=bondorders
            )
        else:
            self._system['templatebond'].append(i=ti, j=tj)

    def bonds(self, subset=None, configuration=None):
        """Returns an iterator over the rows of the bonds.

        Note that these are references to the main atom table.

        Parameters
        ----------
        subset : int = None
            Get the atoms for the subset. Defaults to the 'all/all' subset
            for the configuration given.
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration. Not used if the subset is given.

        Returns
        -------
        sqlite3.Cursor
            A cursor that returns sqlite3.Row objects for the bonds.
        """
        if subset is None:
            subset = self.system.all_subset(configuration)

        columns = []
        for column in self.attributes:
            if column != 'i' and column != 'j':
                columns.append(f'templatebond."{column}"')
        column_defs = ', '.join(columns)
        sql = (
            f"SELECT iatom.atom as i, jatom.atom as j, {column_defs}"
            "   FROM templatebond, subset_atom as iatom, subset_atom as jatom"
            "  WHERE templatebond.i = iatom.templateatom"
            "    AND templatebond.j = jatom.templateatom"
            "    AND iatom.subset = ? and jatom.subset = ?"
        )
        return self.db.execute(sql, (subset, subset))

    def contains_bond(self, key):
        if isinstance(key, self.bond_tuple):
            i = key.i
            j = key.j
        else:
            i, j = key
        # get canonical order
        if i > j:
            j, i = i, j
        sql = (
            "SELECT COUNT(*) FROM templatebond, subset_atom as iatom,"
            "              subset_atom as jatom"
            " WHERE iatom.subset = ? AND jatom.subset = ?"
            "   AND templatebond.i = iatom.templateatom"
            "   AND templatebond.j = jatom.templateatom"
            "   AND iatom.atom = ? AND jatom.atom = ?"
        )
        all_subset = self._system.all_subset()
        self.cursor.execute(sql, (all_subset, all_subset, i, j))
        return self.cursor.fetchone()[0] == 1

    def delete_bond(self, i, j):
        """Remove a bond, if present"""
        # get canonical order
        if i > j:
            j, i = i, j
        sql = (
            "SELECT templatebond.i as i, templatebond.j as j"
            "  FROM templatebond, subset_atom as iatom, subset_atom as jatom"
            " WHERE iatom.subset = ? AND jatom.subset = ?"
            "   AND templatebond.i = iatom.templateatom"
            "   AND templatebond.j = jatom.templateatom"
            "   AND iatom.atom = ? AND jatom.atom = ?"
        )
        all_subset = self._system.all_subset()
        self.cursor.execute(sql, (all_subset, all_subset, i, j))
        row = self.cursor.fetchone()
        if row is not None:
            sql = "DELETE FROM templatebond WHERE i = ? AND j = ?"
            self.cursor.execute(sql, row)

    def get_bond(self, i, j):
        # get canonical order
        if i > j:
            j, i = i, j
        sql = (
            "SELECT templatebond.* FROM templatebond, subset_atom as iatom,"
            "              subset_atom as jatom"
            " WHERE iatom.subset = ? AND jatom.subset = ?"
            "   AND templatebond.i = iatom.templateatom"
            "   AND templatebond.j = jatom.templateatom"
            "   AND iatom.atom = ? AND jatom.atom = ?"
        )
        all_subset = self._system.all_subset()
        cursor = self.db.execute(sql, (all_subset, all_subset, i, j))
        row = cursor.fetchone()
        if row is None:
            raise KeyError(f'No bond from {i} to {j} found')
        return row

    def get_column(self, key, configuration=None):
        """Return a column from the templatebonds table.

        Parameters
        ----------
        key : str
            The column (attribute) to getLogger
        configuration : int = None
            Get the bonds for this configuration, defaults to the current
            configuration.

        Returns
        -------
        column : Column
            The column object requested.
        """
        if key == 'i' or key == 'j':
            all_subset = self._system.all_subset(configuration)
            if key == 'i':
                col = 'iatom.atom'
            else:
                col = 'jatom.atom'
            sql = (
                f"SELECT {col}"
                "  FROM templatebond,"
                "       subset_atom as iatom,"
                "       subset_atom as jatom"
                " WHERE templatebond.i = iatom.templateatom"
                "   AND templatebond.j = jatom.templateatom"
                f"  AND iatom.subset = {all_subset}"
                f"  AND jatom.subset = {all_subset}"
            )
            table = Table(self._system, 'atom')
            return FrozenColumn(table, 'id', sql)
        else:
            all_template = self._system.all_template(configuration)
            sql = (
                f'SELECT templatebond.rowid, templatebond."{key}"'
                "  FROM templatebond,"
                "       templateatom as iatom,"
                "       templateatom as jatom"
                " WHERE templatebond.i = iatom.id"
                "   AND templatebond.j = jatom.id"
                f"  AND iatom.template = {all_template}"
                f"  AND jatom.template = {all_template}"
            )
            table = Table(self._system, 'templatebond')
            return Column(table, key, sql=sql)

    def n_bonds(self, subset: int = None, configuration: int = None) -> int:
        """The number of bonds.

        Parameters
        ----------
        bonds : int = None
            Get the bonds for the subset. Defaults to the 'all/all' subset
            for the configuration given.
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration. Not used if the subset is given.

        Returns
        -------
        int
            Number of bonds
        """
        if subset is None:
            template = self.system.all_template(configuration)
        else:
            template = self.system.subsets.template(subset)

        self.cursor.execute(
            "SELECT COUNT(*) FROM templateatom, templatebond "
            " WHERE templatebond.i = templateatom.id"
            "   AND templateatom.template = ?", (template,)
        )
        return self.cursor.fetchone()[0]

    def remove(self, subset=None, configuration=None):
        """Removes all the bonds in a subset or configuration

        Parameters
        ----------
        subset : int = None
            Get the atoms for the subset. Defaults to the 'all/all' subset
            for the configuration given.
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration. Not used if the subset is given.

        Returns
        -------
        None
        """
        if subset is None:
            subset = self.system.all_subset(configuration)

        sql = (
            "DELETE FROM templatebond"
            " WHERE i in ("
            "     SELECT id FROM templateatom, subset_atom"
            "      WHERE id = templateatom AND subset = ?"
            " ) AND j in ("
            "     SELECT id FROM templateatom, subset_atom"
            "      WHERE id = templateatom AND subset = ?"
            " )"
        )
        self.db.execute(sql, (subset, subset))

    def to_dataframe(self, configuration=None):
        """Return the bonds as a Pandas Dataframe."""
        all_subset = self._system.all_subset(configuration)
        columns = []
        for column in self.attributes:
            if column != 'i' and column != 'j':
                columns.append(f'templatebond."{column}"')
        column_defs = ', '.join(columns)
        sql = (
            "SELECT templatebond.rowid,"
            "       iatom.atom as i,"
            "       jatom.atom as j,"
            f"      {column_defs}"
            "  FROM templatebond, subset_atom as iatom, subset_atom as jatom"
            " WHERE templatebond.i = iatom.templateatom"
            "   AND templatebond.j = jatom.templateatom"
            "   AND iatom.subset = ? and jatom.subset = ?"
        )

        data = {}
        for line in self.cursor.execute(sql, (all_subset, all_subset)):
            data[line[0]] = line[1:]
        columns = [x[0] for x in self.cursor.description[1:]]

        df = pandas.DataFrame.from_dict(data, orient='index', columns=columns)

        return df

        # columns = [*self.attributes]
        # df = pandas.DataFrame.from_records(
        #     self.bonds(configuration), columns=columns
        # )
        # return df

    def _row_factory(self, cursor, row):
        return self.bond_tuple(*row)



# -*- coding: utf-8 -*-

import logging
import math

import numpy

logger = logging.getLogger(__name__)


def cos(value):
    return math.cos(math.radians(value))


def sin(value):
    return math.sin(math.radians(value))


class Cell(object):
    """A class to handle cell parameters and their transformations."""

    def __init__(self, a, b, c, alpha, beta, gamma):
        self._parameters = [a, b, c, alpha, beta, gamma]

    def __getitem__(self, key):
        """Allow [] to access the data!"""
        return self._parameters[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._parameters[key] = value

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._parameters)

    def __len__(self) -> int:
        """The len() command"""
        return len(self._parameters)

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        return self.equal(other, tol=1.0e-12)

    def __repr__(self):
        """The representation of this object"""
        return repr(self._parameters)

    def __str__(self):
        """The pretty string representation of this object"""
        return str(self._parameters)

    @property
    def a(self):
        """The length of the first cell vector."""
        return self._parameters[0]

    @a.setter
    def a(self, value):
        self._parameters[0] = value

    @property
    def b(self):
        """The length of the second cell vector."""
        return self._parameters[1]

    @b.setter
    def b(self, value):
        self._parameters[1] = value

    @property
    def c(self):
        """The length of the third cell vector."""
        return self._parameters[2]

    @c.setter
    def c(self, value):
        self._parameters[2] = value

    @property
    def alpha(self):
        """The angle between b and c."""
        return self._parameters[3]

    @alpha.setter
    def alpha(self, value):
        self._parameters[3] = value

    @property
    def beta(self):
        """The angle between a and c."""
        return self._parameters[4]

    @beta.setter
    def beta(self, value):
        self._parameters[4] = value

    @property
    def gamma(self):
        """The angle between a and b."""
        return self._parameters[5]

    @gamma.setter
    def gamma(self, value):
        self._parameters[5] = value

    @property
    def parameters(self):
        """The cell parameters as a list."""
        return list(self._parameters)

    @parameters.setter
    def parameters(self, value):
        if len(value) != 6:
            raise ValueError('parameters must be of length 6')
        self._parameters = list(value)

    @property
    def volume(self):
        """The volume of the cell."""
        a, b, c, alpha, beta, gamma = self.parameters
        return (
            a * b * c * (1 - cos(alpha)**2 - cos(beta)**2 - cos(gamma)**2) +
            2 * math.sqrt(cos(alpha) * cos(beta) * cos(gamma))
        )

    def equal(self, other, tol=1.0e-6):
        """Check if we are equal to another iterable to within a tolerance.

        Parameters
        ----------
        other : iterable
            The other object to check against
        tol : float = 1.0e-06
            The tolerance for comparing floating point numbers.

        Returns
        -------
        equals : bool
            Boolean indicating whether the two are equal.
        """
        if len(other) != 6:
            return False

        for i, j in zip(self, other):
            if abs(i - j) > tol:
                return False

        return True

    def to_cartesians(self, uvw, as_array=False):
        """Convert fraction coordinates to Cartesians

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        uvw : [N][3*float] or ndarray
            The fractional coordinates.

        Returns
        -------
        xyz : [N][float*3] or ndarray
            The Cartesian coordinates.
        """
        if isinstance(uvw, numpy.ndarray):
            UVW = uvw
        else:
            UVW = numpy.array(uvw)

        T = self.to_cartesians_transform(as_array=True)
        XYZ = UVW @ T

        if as_array:
            return XYZ
        else:
            return XYZ.tolist()

    def to_cartesians_transform(self, as_array=False):
        """Matrix to convert fractional coordinates to Cartesian.

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        as_array : bool = False
            Whether to return a numpy array or Python lists

        Returns
        -------
        transform : [N][float*3] or ndarray
            The transformation matrix
        """
        a, b, c, alpha, beta, gamma = self.parameters

        ca = cos(alpha)
        cb = cos(beta)
        cg = cos(gamma)
        sg = sin(gamma)

        V = a * b * c * math.sqrt(1 - ca**2 - cb**2 - cg**2 + 2 * ca * cb * cg)
        # Transpose of ...
        # [a, b * cg, c * cb],
        # [0, b * sg, c * (ca - cb * cg) / sg],
        # [0, 0, V / (a * b * sg)]
        T = [
                [
                    a,
                    0,
                    0
                ],
                [
                    b * cg,
                    b * sg,
                    0
                ],
                [
                    c * cb,
                    c * (ca - cb * cg) / sg,
                    V / (a * b * sg)
                ]
            ]  # yapf: disable

        if as_array:
            return numpy.array(T)
        else:
            return T

    def to_fractionals(self, xyz, as_array=False):
        """Convert Cartesian coordinates to fractional.

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        xyz : [N][3*float] or ndarray
            The Cartesian coordinates.

        Returns
        -------
        uvw : [N][float*3] or ndarray
            The ractional coordinates.
        """
        if isinstance(xyz, numpy.ndarray):
            XYZ = xyz
        else:
            XYZ = numpy.array(xyz)

        T = self.to_fractionals_transform(as_array=True)
        UVW = XYZ @ T

        if as_array:
            return UVW
        else:
            return UVW.tolist()

    def to_fractionals_transform(self, as_array=False):
        """Matrix to convert Cartesian coordinates to fractional.

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        as_array : bool = False
            Whether to return a numpy array or Python lists

        Returns
        -------
        transform : [N][float*3] or ndarray
            The transformation matrix
        """
        a, b, c, alpha, beta, gamma = self.parameters

        ca = cos(alpha)
        cb = cos(beta)
        cg = cos(gamma)
        sg = sin(gamma)

        V = a * b * c * math.sqrt(1 - ca**2 - cb**2 - cg**2 + 2 * ca * cb * cg)
        # Transpose...
        # [1 / a, -cg / (a * sg), b * c * (ca * cg - cb) / (V * sg)],
        # [0, 1 / (b * sg), a * c * (cb * cg - ca) / (V * sg)],
        # [0, 0, a * b * sg / V]
        T = [
                [
                    1 / a,
                    0,
                    0
                ],
                [
                    -cg / (a * sg),
                    1 / (b * sg),
                    0
                ],
                [
                    b * c * (ca * cg - cb) / (V * sg),
                    a * c * (cb * cg - ca) / (V * sg),
                    a * b * sg / V
                ]
            ]  # yapf: disable

        if as_array:
            return numpy.array(T)
        else:
            return T



# -*- coding: utf-8 -*-

"""A class providing a convenient interface for subsets
"""

from itertools import zip_longest
import logging
from typing import TypeVar, Dict, Any

from molsystem.table import _Table as Table

System_tp = TypeVar("System_tp", "System", None)
Templates_tp = TypeVar("Templates_tp", "_Templates", str, None)

logger = logging.getLogger(__name__)


def grouped(iterable, n):
    "s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,...s3n-1), ..."
    return zip_longest(*[iter(iterable)] * n)


class _Subsets(Table):
    """The Subset class works with the tables controlling subsets.

    See the main documentation of SEAMM for a detailed description of the
    database scheme underlying the system and hence the subsets.. The
    following tables handle subsets:

    """

    def __init__(self, system: System_tp, tablename: str = 'subset') -> None:
        super().__init__(system, tablename)

        self._configuration_subset_table = self.system['configuration_subset']

    def n_subsets(self, configuration=None):
        """The number of subsets for a configuration.

        Parameters
        ----------
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration.

        Returns
        -------
        int
            The number of subsets in the configuration.
        """
        if configuration is None:
            configuration = self.system.current_configuration

        self.cursor.execute(
            f'SELECT COUNT(*) FROM {self.table}, "configuration_subset"'
            '  WHERE id = subset AND configuration = ?', (configuration,)
        )
        result = self.cursor.fetchone()[0]
        return result

    def append(
        self,
        n: int = None,
        configuration: int = None,
        **kwargs: Dict[str, Any]
    ) -> None:
        """Append one or more rows

        The keywords are the names of attributes and the value to use.
        The default value for any attributes not given is used unless it is
        'None' in which case an error is thrown. It is an error if there is not
        an exisiting attribute corresponding to any given as arguments.

        Parameters
        ----------
        n : int = None
            The number of rows to create, defaults to the number of items
            in the longest attribute given.
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration.
        kwargs :
            any number <attribute name> = <value> keyword arguments giving
            existing attributes and values.

        Returns
        -------
        [int]
            The ids of the created rows.
        """
        if configuration is None:
            configuration = self.system.current_configuration

        ids = super().append(n, **kwargs)

        # and link to the configuration
        self._configuration_subset_table.append(
            configuration=configuration, subset=ids
        )

        return ids

    def delete(self, ids, configuration=None):
        """Remove one or more subsets

        Parameters
        ----------
        ids : [int]
            The subsets to delete.
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration.

        Returns
        -------
        None
        """
        if ids == 'all':
            if configuration is None:
                configuration = self.system.current_configuration
            sql = (
                f'DELETE FROM {self.table}, "configuration_subset"'
                '  WHERE id = subset and configuration = ?'
            )
            self.db.execute(sql, (configuration,))
        else:
            if isinstance(ids, int):
                self.db.execute(
                    f"DELETE FROM {self.table} WHERE id = ?", (ids,)
                )
            else:
                self.db.executemany(
                    f"DELETE FROM {self.table} WHERE id = ?", ids
                )

    def create(
        self, template, configuration=None, atoms=None, templateatoms=None
    ):
        """Create a subset given a template and optionally atoms.

        Parameters
        ----------
        template : int
            The template for the subset
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration.
        atoms : [int] = None
            Optional list of atom ids to connect to the subset.
        templateatoms : [int] = None
            Optional list of template atoms to connect to the atoms
            in the subset.

        Returns
        -------
        int
            The id of the subset.
        """
        sid = self.append(template=template, configuration=configuration)[0]

        if atoms is not None:
            sa = self.system['subset_atom']
            if templateatoms is not None:
                sa.append(subset=sid, atom=atoms, templateatom=templateatoms)
            else:
                sa.append(subset=sid, atom=atoms)

        return sid

    def find(self, template, configuration=None):
        """Find subsets given a template.

        Parameters
        ----------
        template : int
            The template for the subset
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration.

        Returns
        -------
        [int]
            The ids of the subsets, and empty list if there are none.
        """
        if configuration is None:
            configuration = self.system.current_configuration

        result = []
        sql = (
            f'SELECT id FROM {self.table}, "configuration_subset"'
            '  WHERE id = subset AND configuration = ? AND template = ?'
        )
        for row in self.db.execute(sql, (configuration, template)):
            result.append(row['id'])

        return result

    def template(self, sid):
        """The template for the given subset.

        Parameters
        ----------
        sid : int
            The id of the subset.

        Returns
        -------
        int
            The id of the associated template
        """
        self.cursor.execute(
            f'SELECT template FROM {self.table} WHERE id = ?', (sid,)
        )
        return self.cursor.fetchone()[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the system class."""

import pprint

import pytest  # noqa: F401


def test_construction(system):
    """Simplest test that we can make a System object"""
    assert str(type(system)) == "<class 'molsystem.system._System'>"


def test_version_empty(system):
    """Simplest test that we can make a System object"""
    assert system.version == 0


def test_create_table(system):
    """Test that we can create a table in the system."""
    with system as tmp:
        table = tmp.create_table('table1')
        # Create a column so the table exists
        table.add_attribute('data')
    assert system.version == 1
    assert 'table1' in system


def test_create_table_(system):
    """Test that we can create a table using [] syntax."""
    with system as tmp:
        table = tmp['table1']
        # Create a column so the table exists
        table.add_attribute('data')
    assert system.version == 1
    assert 'table1' in system


def test_delete_table(system_with_two_tables):
    """Test that we can delete a table."""
    system = system_with_two_tables
    with system as tmp:
        del tmp['table1']
    assert system.version == 1
    assert 'table2' in system
    assert 'table1' not in system


def test_elements(system):
    """Test that we can access the element table"""
    table = system['element']
    assert table.n_rows == 118


def test_formula(AceticAcid):
    """Test the formula generation."""
    formula, empirical_formula, Z = AceticAcid.formula()
    assert ''.join(formula) == 'C2H4O2'
    assert ''.join(empirical_formula) == 'CH2O'
    assert Z == 2


def test_formula_periodic(vanadium):
    """Test the formula generation."""
    formula, empirical_formula, Z = vanadium.formula()
    assert ''.join(formula) == 'V2'
    assert ''.join(empirical_formula) == 'V'
    assert Z == 2


def test_clear(CH3COOH_3H2O):
    """Test making subsets for the molecules."""
    result = [2, 3, 4, 5]

    system = CH3COOH_3H2O
    sids = system.create_molecule_subsets()
    if sids != result:
        pprint.pprint(sids)
    assert sids == result

    system.clear()

    assert system.n_atoms() == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Fixtures for testing the 'molsystem' package."""
import math

import pytest

from molsystem.systems import Systems


def pytest_addoption(parser):
    parser.addoption(
        "--run-timing",
        action="store_true",
        default=False,
        help="run timing tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "timing: mark test as timing to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run-timing"):
        # --run-timing given in cli: do not skip timing tests
        return
    skip_timing = pytest.mark.skip(reason="need --run-timing option to run")
    for item in items:
        if "timing" in item.keywords:
            item.add_marker(skip_timing)


def mk_table(system, name='table1'):
    """Create a table with some attributes."""
    table = system.create_table(name)
    table.add_attribute('atno', coltype='int', default=-1)
    for column in ('x', 'y', 'z'):
        table.add_attribute(column, coltype='float')
    return table


@pytest.fixture()
def atoms(system):
    """An empty atoms table."""
    return system.atoms


@pytest.fixture()
def system():
    systems = Systems()
    system = systems.create_system('seamm', temporary=True)

    yield system

    del systems['seamm']


@pytest.fixture()
def two_systems():
    systems = Systems()

    system1 = systems.create_system('seamm1', temporary=True)
    system2 = systems.create_system('seamm2', temporary=True)

    yield system1, system2

    del systems['seamm1']
    del systems['seamm2']


@pytest.fixture()
def simple_table(system):
    return mk_table(system)


@pytest.fixture()
def two_tables(system):
    return mk_table(system, 'table1'), mk_table(system, 'table2')


@pytest.fixture()
def system_with_two_tables(system):
    mk_table(system, 'table1')
    mk_table(system, 'table2')
    return system


@pytest.fixture()
def two_tables_in_two_systems(two_systems):
    system, system2 = two_systems
    return mk_table(system, 'table1'), mk_table(system2, 'table2')


@pytest.fixture()
def bonds(system):
    """An empty system object
    """
    return system


@pytest.fixture()
def templates(system):
    """A system with a template for water."""
    templates = system['template']
    atoms = system['templateatom']
    bonds = system['templatebond']

    # TIP3P
    r0 = 0.9572
    theta0 = 104.52

    # H locations are ±x, 0, z
    x = r0 * math.sin(math.radians(theta0 / 2))
    z = r0 * math.cos(math.radians(theta0 / 2))

    X = [0.0, x, -x]
    Y = [0.0, 0.0, 0.0]
    Z = [0.0, z, z]

    atno = [8, 1, 1]
    name = ['O', 'H1', 'H2']
    i_atom = [0, 0]
    j_atom = [1, 2]

    tid = templates.append(name='H2O', type='molecule')[0]
    templates.current_template = tid

    ids = atoms.append(atno=atno, x=X, y=Y, z=Z, name=name)

    i = [ids[x] for x in i_atom]
    j = [ids[x] for x in j_atom]

    bonds.append(i=j, j=i)  # flipped on purpose so code orders.

    # Acetic acid

    # yapf: disable
    #       C       H        H        H        C        =O      O        H
    X = [ 1.0797, 0.5782,  0.7209,  0.7052,  0.5713, -0.1323, 0.9757,  2.1724]  # noqa: E221, E501, E201
    Y = [ 0.0181, 3.1376, -0.6736, -0.3143,  1.3899,  1.7142, 2.2970,  0.0161]  # noqa: E221, E501, E201
    Z = [-0.0184, 0.2813, -0.7859,  0.9529, -0.3161, -1.2568, 0.5919, -0.0306]  # noqa: E221, E501
    symbol = ['C', 'H', 'H', 'H', 'C', 'O', 'O', 'H']
    name = ['C1', 'H1', 'H2', 'H3', 'C', 'O', 'OH', 'H']

    #       C  H  H  H  C =O  O  H
    i_atom = [0, 0, 0, 0, 4, 4, 6]
    j_atom = [1, 2, 3, 4, 5, 6, 7]
    order =  [1, 1, 1, 1, 2, 1, 1]  # noqa: E222
    # yapf: enable

    tid = templates.append(name='acetic acid', type='molecule')[0]
    templates.current_template = tid

    ids = atoms.append(symbol=symbol, x=X, y=Y, z=Z, name=name)

    i = [ids[x] for x in i_atom]
    j = [ids[x] for x in j_atom]

    bonds.append(i=i, j=j, bondorder=order)

    return system


@pytest.fixture()
def AceticAcid(system):
    """An system object for an acetic acid molecule
    """
    # yapf: disable
    #       C       H        H        H        C        =O      O        H
    x = [ 1.0797, 0.5782,  0.7209,  0.7052,  0.5713, -0.1323, 0.9757,  2.1724]  # noqa: E221, E501, E201
    y = [ 0.0181, 3.1376, -0.6736, -0.3143,  1.3899,  1.7142, 2.2970,  0.0161]  # noqa: E221, E501, E201
    z = [-0.0184, 0.2813, -0.7859,  0.9529, -0.3161, -1.2568, 0.5919, -0.0306]  # noqa: E221, E501
    atno = [6, 1, 1, 1, 6, 8, 8, 1]  # noqa: E221

    #       C  H  H  H  C =O  O  H
    i_atom = [0, 0, 0, 0, 4, 4, 6]
    j_atom = [1, 2, 3, 4, 5, 6, 7]
    order =  [1, 1, 1, 1, 2, 1, 1]  # noqa: E222
    # yapf: enable

    system.name = 'acetic acid'
    ids = system['atoms'].append(x=x, y=y, z=z, atno=atno)

    i = [ids[x] for x in i_atom]
    j = [ids[x] for x in j_atom]

    system['bonds'].append(i=i, j=j, bondorder=order)

    return system


@pytest.fixture
def disordered(AceticAcid):
    """Two acetic acid molecules with different atom order"""
    # yapf: disable
    #       C       H        H        H        C        =O      O        H
    x = [ 1.0797, 0.5782,  0.7209,  0.7052,  0.5713, -0.1323, 0.9757,  2.1724]  # noqa: E221, E501, E201
    y = [ 0.0181, 3.1376, -0.6736, -0.3143,  1.3899,  1.7142, 2.2970,  0.0161]  # noqa: E221, E501, E201
    z = [-0.0184, 0.2813, -0.7859,  0.9529, -0.3161, -1.2568, 0.5919, -0.0306]  # noqa: E221, E501
    atno = [6, 1, 1, 1, 6, 8, 8, 1]  # noqa: E221

    #       C  H  H  H  C =O  O  H
    i_atom = [0, 0, 0, 0, 4, 4, 6]
    j_atom = [1, 2, 3, 4, 5, 6, 7]
    order =  [1, 1, 1, 1, 2, 1, 1]  # noqa: E222
    # yapf: enable

    system = AceticAcid
    ids = system['atoms'].append(
        x=[*reversed(x)],
        y=[*reversed(y)],
        z=[*reversed(z)],
        atno=[*reversed(atno)]
    )

    i = [ids[7 - x] for x in i_atom]
    j = [ids[7 - x] for x in j_atom]

    system['bonds'].append(i=i, j=j, bondorder=order)

    return system


@pytest.fixture()
def vanadium(system):
    """BCC vanadium crystal, without symmetry."""
    system.name = 'BCC Vanadium'
    system.periodicity = 3
    system.coordinate_system = 'fractional'
    system.cell.set_cell(3.03, 3.03, 3.03, 90, 90, 90)
    system.atoms.append(x=[0.0, 0.5], y=[0.0, 0.5], z=[0.0, 0.5], symbol='V')
    return system


@pytest.fixture()
def CH3COOH_3H2O(AceticAcid):
    """System with acetic acid and 3 water molecules"""
    system = AceticAcid

    # TIP3P
    r0 = 0.9572
    theta0 = 104.52

    # H locations are ±x, 0, z
    x = r0 * math.sin(math.radians(theta0 / 2))
    z = r0 * math.cos(math.radians(theta0 / 2))

    X = [0.0, x, -x]
    Z = [0.0, z, z]

    atno = [8, 1, 1]
    i_atom = [0, 0]
    j_atom = [1, 2]

    system = AceticAcid
    system.name = 'acetic acid with 3 waters'

    for no in range(1, 4):
        ids = system['atoms'].append(x=X, y=no * 5.0, z=Z, atno=atno)

        i = [ids[x] for x in i_atom]
        j = [ids[x] for x in j_atom]

        system['bonds'].append(i=i, j=j)

    return system



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pytest  # noqa: F401

import molsystem  # noqa: F401
"""Tests for `molsystem` package."""

x = [1.0, 2.0, 3.0]
y = [4.0, 5.0, 6.0]
z = [7.0, 8.0, 9.0]
atno = [8, 1, 1]


def test_construction(system):
    """Simplest test that we can make an Atoms object"""
    atoms = system.atoms
    assert str(type(atoms)) == "<class 'molsystem.atoms._Atoms'>"


def test_keys(atoms):
    """Test the default keys in an Atoms object"""
    result = ['atno', 'configuration', 'id', 'x', 'y', 'z']

    assert sorted([*atoms.keys()]) == result


def test_n_atoms_empty(atoms):
    """Test how many atoms an empty object has"""
    assert atoms.n_atoms() == 0


def test_append_one(atoms):
    """Test adding one atom"""
    with atoms as tmp:
        tmp.append(x=1.0, y=2.0, z=3.0, atno=[6])
    assert atoms.n_atoms() == 1


def test_append_several(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms.n_atoms() == 3


def test_append_several_scalar(atoms):
    """Test adding several atoms with some scalar values"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=0.0, atno=6)
    assert (
        atoms['x'] == x and atoms['y'] == y and
        atoms['z'] == [0.0, 0.0, 0.0] and atoms['atno'] == [6, 6, 6]
    )


def test_append_using_default(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z)

    assert (atoms.n_atoms() == 3 and atoms['atno'] == [None, None, None])


def test_append_error(atoms):
    """Test adding atoms with an error"""
    atoms.configuration = 1
    with pytest.raises(KeyError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno, bad=99)
    assert atoms.n_atoms() == 0 and atoms.version == 0
    assert (str(e.value) == '\'"bad" is not an attribute of the atoms.\'')


def test_add_attribute(atoms):
    """Test adding an attribute"""
    result = ['atno', 'configuration', 'id', 'name', 'x', 'y', 'z']
    with atoms as tmp:
        tmp.add_attribute('name')
    assert sorted([*atoms.keys()]) == result


def test_add_duplicate_attribute(atoms):
    """Test duplicate adding an attribute"""
    result = ['atno', 'configuration', 'id', 'name', 'x', 'y', 'z']
    with atoms as tmp:
        tmp.add_attribute('name')
    with pytest.raises(RuntimeError) as e:
        with atoms as tmp:
            ver = tmp.version
            tmp.add_attribute('name')
    assert (
        sorted([*atoms.keys()]) == result and ver == 1 and
        atoms.version == 1 and
        str(e.value) == "_Table attribute 'name' is already defined!"
    )


def test_add_attribute_with_values(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute('name', values=['H1', 'O', 'H2'])
    assert np.array_equal(atoms['name'], ['H1', 'O', 'H2'])


def test_add_attribute_with_one_value(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute('name', values=['H1'])
    assert np.array_equal(atoms['name'], ['H1', 'H1', 'H1'])


def test_add_attribute_with_one_value_not_list(atoms):
    """Test adding an attribute using a scalar value"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute('spin', coltype='int', default=0, values=1)
    assert atoms['spin'] == [1, 1, 1]


def test_add_attribute_with_wrong_number_of_values(atoms):
    """Test adding an attribute using a the wrong number of values"""
    with pytest.raises(IndexError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno)
            tmp.add_attribute('spin', coltype='int', default=0, values=[1, 2])
    assert (
        atoms.n_atoms() == 0 and str(e.value) == (
            "The number of values given, "
            '2, must be either 1, or the number of rows in "atom": 3'
        )
    )


def test_get_attribute_by_index(atoms):
    """Get a single value of an attribute by index"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms['atno'][1] == 1


def test_get_attribute_by_slice(atoms):
    """Get several values using a slice"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms['atno'][::2] == [8, 1]


def test_getting_x(atoms):
    """Test getting the x column of coordinates"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms['x'] == [1.0, 2.0, 3.0]


def test_contains(atoms):
    """Test the __contains__ or 'in' functionalty"""
    assert 'x' in atoms


def test_not_in(atoms):
    """Test the __contains__ or 'in' functionalty"""
    assert 'abc' not in atoms


def test_deleting_column(atoms):
    """Test deleting a column"""
    with atoms as tmp:
        del tmp['atno']
    assert sorted([*atoms.keys()]) == ['configuration', 'id', 'x', 'y', 'z']


def test_set_column(atoms):
    """Test setting a column using a scalar"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms as tmp:
        tmp['atno'] = 10

    assert (
        atoms.n_atoms() == 3 and atoms.version == 2 and
        atoms['atno'] == [10, 10, 10]
    )


def test_set_column_with_array(atoms):
    """Test setting a column using an array"""
    values = [10.0, 11.0, 12.0]

    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms as tmp:
        tmp['x'] = values

    assert (
        atoms.n_atoms() == 3 and atoms.version == 2 and atoms['x'] == values
    )


def test_append_error_no_coordinates(atoms):
    """Test adding atoms without coordinates, raising an error"""
    with atoms as tmp:
        tmp.append(y=y, z=z, atno=atno)

    assert (atoms.n_atoms() == 3 and atoms['x'] == [None, None, None])


def test_append_error_invalid_column(atoms):
    """Test adding atoms without coordinates, raising an error"""
    with pytest.raises(KeyError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno, junk=99)
    assert (
        atoms.n_atoms() == 0 and
        str(e.value) == '\'"junk" is not an attribute of the atoms.\''
    )


def test_append_error_invalid_length(atoms):
    """Test adding atoms without coordinates, raising an error"""
    with pytest.raises(IndexError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=[3, 4])
    assert (
        atoms.n_atoms() == 0 and str(e.value) == (
            'key "atno" has the wrong number of values, '
            '2. Should be 1 or the number of atoms (3).'
        )
    )


def test_add_attribute_with_no_default(atoms):
    """Test adding an attribute with no default, then several atoms"""
    with atoms as tmp:
        tmp.add_attribute('new', coltype='float')
        tmp.append(x=x, y=y, z=z, atno=atno, new=[-1, -2, -3])

    assert (atoms.n_atoms() == 3 and atoms['new'] == [-1, -2, -3])


def test_equality(two_systems):
    """Test whether two Atoms objects are equal"""
    system1, system2 = two_systems
    atoms1 = system1['atoms']
    atoms2 = system2['atoms']

    with atoms1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms2 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    assert atoms1.n_atoms() == 3 and atoms1 == atoms2


def test_inequality(two_systems):
    """Test whether two Atoms objects are equal"""
    system1, system2 = two_systems
    atoms1 = system1['atoms']
    atoms2 = system2['atoms']

    with atoms1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms2 as tmp:
        atno2 = list(atno)
        atno2[2] = 13
        tmp.append(x=x, y=y, z=z, atno=atno2)

    assert atoms1.n_atoms() == 3 and atoms1 != atoms2


def test_str(atoms):
    """Test string representation of atoms object"""
    string_rep = """\
   atno  configuration    x    y    z
1     8              1  1.0  4.0  7.0
2     1              1  2.0  5.0  8.0
3     1              1  3.0  6.0  9.0"""  # noqa: W291

    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    if str(atoms) != string_rep:
        print(str(atoms))

    assert str(atoms) == string_rep


def test_repr(atoms):
    """Test representation of atoms object"""
    string_rep = """\
   atno  configuration    x    y    z
1     8              1  1.0  4.0  7.0
2     1              1  2.0  5.0  8.0
3     1              1  3.0  6.0  9.0"""  # noqa: W291

    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    if repr(atoms) != string_rep:
        print(repr(atoms))

    assert repr(atoms) == string_rep


def test_equals(atoms):
    """Test shallow copy"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    atoms2 = atoms

    assert atoms == atoms2


def test_equals_with_change(atoms):
    """Test (shallow) copy and changing copy"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    atoms2 = atoms
    atoms2['atno'][0] = 22

    assert atoms == atoms2



# -*- coding: utf-8 -*-

import logging
from typing import Any, Dict, TypeVar

from molsystem.table import _Table as Table
"""A dictionary-like object for holding bonds

Based on tables in an SQLite database.
"""

System_tp = TypeVar("System_tp", "System", "Bonds", None)

logger = logging.getLogger(__name__)


class _Templatebonds(Table):
    """The Bonds class holds arrays of attributes describing bonds

    Three attributes are required: the atoms 'i' and 'j' of the bond, and the
    bond order, which defaults to a single bond.

    Since there are two possible representations for a bond, i-j and j-i,
    the bonds are stored with i < j in order to make the indexing unique.

    In order to handle changes in bonding (and numbers of atoms) the system
    class has several tables covering the bonds. See the main documentation for
    a more detailed description. The key tables are:

    template -- a simple list of templates
    subset -- an instantiation of a template, connected with one or more
              configurations of the system.
    subset_atom -- connects the subset to the atoms, and optionally connects
                   the atom to a template atom.
    templateatom -- holds the atoms in a template, if present.
    templatebond -- holds the bonds between the template atoms.

    Other attributes can be created, either from a list of predefined
    ones or by specifying the metadata required of an attribute. Attributes can
    also be removed. See the method 'add_attribute' for more detail.

    Bonds can be added ('append') or removed ('delete').
    """

    def __init__(self, system: System_tp, table: str = 'templatebond') -> None:

        super().__init__(system, table)

        self._templates = system['template']
        self._templateatoms = system['templateatom']

    def append(self, template: int = None, **kwargs: Dict[str, Any]) -> None:
        """Append one or more bonds

        The keys give the field for the data. If an existing field is not
        mentioned, then the default value is used. It is an error if there is
        not a field corrresponding to a key.

        Parameters
        ----------
        template : int = None
            The template to add the bonds to.
        kwargs : Dict[str, Any]
            The attributes and their values for the bonds to append.

        Returns
        -------
        None
        """
        # Check keys and lengths of added bonds
        if 'i' not in kwargs or 'j' not in kwargs:
            raise KeyError("The atoms i & j are required!")

        if template is None:
            template = self._templates.current_template

        # How many new rows there are
        n_rows, lengths = self._get_n_rows(**kwargs)

        i = kwargs.pop('i')
        j = kwargs.pop('j')

        # Need lists for the zip below to work.
        if lengths['i'] == 0:
            i = [i] * n_rows
        if lengths['i'] == 1 and n_rows > 1:
            i = [i[0]] * n_rows
        if lengths['j'] == 0:
            j = [j] * n_rows
        if lengths['j'] == 1 and n_rows > 1:
            j = [j[0]] * n_rows

        # The list of atom ids in this template, so that we can check the atoms
        ids = self._templateatoms.atom_ids(template=template)

        i2 = []
        j2 = []
        for i_, j_ in zip(i, j):
            if not isinstance(i_, int) or not isinstance(j_, int):
                raise TypeError(
                    f"'i={i_}' and 'j={j_}', the atom indices, must be "
                    "integers"
                )
            if i_ not in ids:
                raise ValueError(f'Atom i ({i_}) is not in the template.')
            if j_ not in ids:
                raise ValueError(f'Atom j ({j_}) is not in the template.')
            # Ensure that i < j
            if i_ < j_:
                i2.append(i_)
                j2.append(j_)
            else:
                i2.append(j_)
                j2.append(i_)

        super().append(i=i2, j=j2, **kwargs)

    def bonds(self, template: int = None) -> int:
        """The an iterator over the bonds in the template

        Parameters
        ----------
        template : int = None
            The template to use, by default the current template.

        Returns
        -------
        n_bonds : int
            The number of bonds in the template
        """
        if template is None:
            template = self._templates.current_template

        return self.db.execute(
            f'SELECT * FROM {self._table}'
            '  WHERE i IN (SELECT id FROM templateatom WHERE template = ?)',
            (template,)
        )

    def n_bonds(self, template: int = None) -> int:
        """The number of bonds in the template

        Parameters
        ----------
        template : int = None
            The template to use, by default the current template.

        Returns
        -------
        n_bonds : int
            The number of bonds in the template
        """
        if template is None:
            template = self._templates.current_template

        self.cursor.execute(
            f'SELECT COUNT(*) FROM {self._table}'
            '  WHERE i IN (SELECT id FROM templateatom WHERE template = ?)'
            '    AND j IN (SELECT id FROM templateatom WHERE template = ?)',
            (template, template)
        )
        return self.cursor.fetchone()[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401

import molsystem  # noqa: F401
"""Tests for `molsystem` package."""

x = [1.0, 2.0, 3.0]
y = [4.0, 5.0, 6.0]
z = [7.0, 8.0, 9.0]
atno = [8, 1, 1]


def test_construction(system):
    """Simplest test that we can make a new template"""
    templates = system['template']
    templates.append(name='H2O', type='molecule')
    assert str(type(templates)) == "<class 'molsystem.template._Template'>"


def test_set_current_template(system):
    """Test that we can set the current template."""
    templates = system['template']
    tid = templates.append(name='H2O', type='molecule')[0]
    templates.current_template = tid
    assert templates.n_rows == 2 and templates.current_template == tid


def test_adding_atoms(system):
    """Test that we can add atoms to a new template."""
    templates = system['template']
    tid = templates.append(name='H2O', type='molecule')[0]
    templates.current_template = tid

    atoms = system['templateatom']
    atoms.append(atno=atno, x=x, y=y, z=z)
    assert atoms.n_atoms == 3


def test_template_str(templates):
    """Print the water template."""
    atom_string = """\
   template name  atno        x    y         z
1         2    O     8  0.00000  0.0  0.000000
2         2   H1     1  0.75695  0.0  0.585882
3         2   H2     1 -0.75695  0.0  0.585882"""

    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    atoms = system['templateatoms']
    if str(atoms) != atom_string:
        print(atoms)
    assert str(atoms) == atom_string


def test_template_n_atoms(templates):
    """Get the number of atoms in acetic acid template."""
    system = templates
    templates = system['template']
    templates.set_current_template('acetic acid', type_='molecule')
    atoms = system['templateatoms']
    assert atoms.n_atoms == 8


def test_invalid_key(templates):
    """Test the error trying to get a non-existant template."""
    system = templates
    templates = system['template']
    with pytest.raises(KeyError) as e:
        templates.current_template = 99
    assert str(e.value) == '''"Template '99' does not exist."'''


def test_invalid_key_2(templates):
    """Test the error trying to get a non-existant template."""
    answer = '''"There is no template 'junk' of type 'molecule'."'''
    system = templates
    templates = system['template']
    with pytest.raises(KeyError) as e:
        templates.set_current_template('junk', type_='molecule')
    assert str(e.value) == answer


def test_get_column_x(templates):
    """Test getting a column from the atoms in a template."""
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    atoms = system['templateatom']
    x = atoms['x']
    assert x.equal([0.0, 0.75695, -0.75695], 1.0e-05)


def test_get_column_atno(templates):
    """Test getting a column from the atoms in a template."""
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    atoms = system['templateatom']
    atno = atoms['atno']
    assert atno == [8, 1, 1]


def test_get_column_error(templates):
    """Test getting a column that doesn't exist."""
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    atoms = system['templateatom']
    with pytest.raises(KeyError) as e:
        junk = atoms['junk']  # noqa: F841
    assert str(e.value) == '''"'junk' not in template atoms"'''


def test_select_atoms(templates):
    """Select just some atoms in the template."""
    system = templates
    templates = system['template']
    templates.set_current_template('acetic acid', type_='molecule')
    atoms = system['templateatom']
    names = []
    x = []
    for row in atoms.atoms('atno', '==', 8):
        names.append(row['name'])
        x.append(row['x'])
    assert names == ['O', 'OH']
    assert x == [-0.1323, 0.9757]


def test_select_atoms_given_template(templates):
    """Select just some atoms in the given template (H2O)."""
    system = templates
    templates = system['template']
    atoms = system['templateatom']
    atoms.current_template = 3  # Acetic acid
    names = []
    x = []
    for row in atoms.atoms('atno', '=', 8, template=2):
        names.append(row['name'])
        x.append(row['x'])
    assert names == ['O']
    assert x == [0.0]


def test_atom_ids(templates):
    """Test getting the atom ids for the current template."""
    system = templates
    templates = system['template']
    templates.set_current_template('acetic acid', type_='molecule')
    atoms = system['templateatom']
    ids = atoms.atom_ids()
    assert ids == [*range(4, 12)]


def test_coordinate_system(templates):
    """Test getting and settign the coordinate system."""
    system = templates
    templates = system['template']
    templates.set_current_template('acetic acid', type_='molecule')
    atoms = system['templateatom']
    assert atoms.coordinate_system == 'Cartesian'

    with pytest.raises(RuntimeError) as e:
        atoms.coordinate_system = 'fractional'
    assert str(e.value) == 'Templates can only use Cartesian coordinates.'


def test_append_just_coordinates(templates):
    """Test appending just the coordinates."""
    system = templates
    templates = system['template']
    templates.set_current_template('acetic acid', type_='molecule')
    atoms = system['templateatom']
    atoms.append(x=[1.0, 2.0, 3.0], y=9.0)
    assert atoms.n_atoms == 11


def test_template_n_bonds(templates):
    """Get the number of bonds in acetic acid template."""
    system = templates
    templates = system['template']
    templates.set_current_template('acetic acid', type_='molecule')
    bonds = system['templatebond']
    assert bonds.n_bonds() == 7


def test_template_n_bonds_given_template(templates):
    """Get the number of bonds in water template."""
    system = templates
    templates = system['template']
    templates.set_current_template('acetic acid', type_='molecule')
    bonds = system['templatebond']
    assert bonds.n_bonds(template=2) == 2


def test_append_bonds_error(templates):
    """Append a bond with an invalid atom id type"""
    answer = "'i=a' and 'j=2', the atom indices, must be integers"
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    bonds = system['templatebond']
    with pytest.raises(TypeError) as e:
        bonds.append(i=['a'], j=[2])
    assert str(e.value) == answer


def test_append_bonds_scalar(templates):
    """Get the number of bonds in water template."""
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    bonds = system['templatebond']
    bonds.append(i=1, j=2)
    assert bonds.n_bonds() == 3


def test_append_bonds_invalid_atom_error(templates):
    """Append a bond with an invalid atom id"""
    answer = "Atom i (5) is not in the template."
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    bonds = system['templatebond']
    with pytest.raises(ValueError) as e:
        bonds.append(i=5, j=[2])
    assert str(e.value) == answer


def test_append_bonds_invalid_atom_j_error(templates):
    """Append a bond with an invalid atom id"""
    answer = "Atom j (5) is not in the template."
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    bonds = system['templatebond']
    with pytest.raises(ValueError) as e:
        bonds.append(i=1, j=5)
    assert str(e.value) == answer


def test_append_bonds_missing_j_error(templates):
    """Append a bond missing the second atom"""
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    bonds = system['templatebond']
    with pytest.raises(KeyError) as e:
        bonds.append(i=1)
    assert str(e.value) == "'The atoms i & j are required!'"


def test_append_bonds_scalar_i(templates):
    """Append a bond with one i atom and several j's"""
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    bonds = system['templatebond']
    bonds.append(i=[1], j=[2, 3])
    assert bonds.n_bonds() == 4


def test_append_bonds_scalar_j(templates):
    """Append a bond with one j atom and several i's"""
    system = templates
    templates = system['template']
    templates.set_current_template('H2O', type_='molecule')
    bonds = system['templatebond']
    bonds.append(i=[2, 3], j=[1])
    assert bonds.n_bonds() == 4



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `cell` in the `molsystem` package."""

import pytest  # noqa: F401


def test_construction(system):
    """Simplest test that we can make a CellParameters object"""
    cell = system['cell']
    assert str(
        type(cell)
    ) == "<class 'molsystem.cell_parameters._CellParameters'>"


def test_periodic_system(system):
    """Test making a simple periodic system."""
    system.periodicity = 3
    system.coordinate_system = 'fractional'
    system.cell.set_cell(3.03, 3.03, 3.03, 90, 90, 90)
    system.atoms.append(x=[0.0, 0.5], y=[0.0, 0.5], z=[0.0, 0.5], symbol='V')
    assert system.n_atoms() == 2 and system.version == 0


def test_periodic_context(system):
    """Test using context."""
    with system as tmp:
        tmp.periodicity = 3
        tmp.coordinate_system = 'fractional'
        tmp.cell.set_cell(3.03, 3.03, 3.03, 90, 90, 90)
        tmp.atoms.append(x=[0.0, 0.5], y=[0.0, 0.5], z=[0.0, 0.5], symbol='V')
    assert system.n_atoms() == 2 and system.version == 1


def test_cell_object(vanadium):
    """Test getting the cell as a Cell object."""
    cell = vanadium.cell.cell()
    assert str(type(cell)) == "<class 'molsystem.cell.Cell'>"
    assert cell == [3.03, 3.03, 3.03, 90.0, 90.0, 90.0]



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding templates


"""

import logging
from typing import TypeVar

from molsystem.table import _Table as Table

System_tp = TypeVar("System_tp", "System", None)
Templates_tp = TypeVar("Templates_tp", "_Templates", str, None)

logger = logging.getLogger(__name__)


class _Template(Table):
    """The Templates class works with SQLite tables describing templates

    See the main documentation of SEAMM for a detailed description of the
    database scheme underlying the system and hence these templates. The
    following tables handle templates:

    template -- A simple table of all of the templates, giving a name and type.

    templateatom -- An optional list of template atoms associated with a
        template.

    templatecoordinates -- An optional set of coordinates for the template
        atoms. These are in Cartesian coordinates since the templates are
        molecular, not periodic, in nature.

    templatebond -- Bonds linking template atoms i & j if present.

    Templates can be added ('append') or removed ('delete').
    """

    def __init__(self, system: System_tp, tablename: str = 'template') -> None:
        super().__init__(system, tablename)

        self._current_template = 1

    @property
    def current_template(self):
        """The template that is the default for the moment."""
        return self._current_template

    @current_template.setter
    def current_template(self, value):
        self.cursor.execute(
            f'SELECT COUNT(*) FROM {self.table} WHERE id = ?', (value,)
        )
        if self.cursor.fetchone()[0] == 0:
            raise KeyError(f"Template '{value}' does not exist.")
        self._current_template = value

    def append_to_system(
        self,
        template,
        n_copies=1,
        coordinates=None,
        configuration=None,
        create_subsets=True
    ):
        """ Append one or more copies of a template to the system.

        Parameters
        ----------

        """
        pass

    def create(self, name, type_='general', atnos=None, bonds=None):
        """Create a new template.

        Parameters
        ----------
        name : str
            The name of the template.
        type_ : str = 'general'
            The type of template, e.g. 'all', 'molecule', 'residue'
        atnos : [int] = None
            The atomic numbers of the template atoms (optional)
        bond : [(int, int, int)]
            The bonds as (i, j, order) (optional)

        Returns
        -------
        int
            The template id.
        """
        if self.exists(name, type_=type_):
            raise KeyError(f"The template '{name}' of type '{type_}' exists.")

        tid = self.append(name=name, type=type_)[0]

        if atnos is not None:
            tatom_ids = self.system.templateatoms.append(
                atno=atnos, template=tid
            )

            if bonds is not None:
                iatoms = []
                jatoms = []
                orders = []
                for i, j, order in bonds:
                    iatoms.append(tatom_ids[i])
                    jatoms.append(tatom_ids[j])
                    orders.append(order)
                self.system.templatebonds.append(
                    template=tid, i=iatoms, j=jatoms, bondorder=orders
                )
        return tid

    def exists(self, name, type_='general'):
        """Return if the template exists given the name and type.

        Parameters
        ----------
        name : str
            The name of the template.
        type_ : str = 'general'
            The type of template, e.g. 'all', 'molecule', 'residue'

        Returns
        -------
        bool
            True if it exists; False otherwise.
        """
        self.cursor.execute(
            f'SELECT COUNT(*) FROM {self.table} WHERE "name" = ? '
            'AND "type" = ?', (name, type_)
        )
        row = self.cursor.fetchone()
        return row[0] == 1

    def find(self, name, type_='general', create=False):
        """Find a single template given the name and typ.

        Parameters
        ----------
        name : str
            The name of the template.
        type_ : str = 'general'
            The type of template, e.g. 'all', 'molecule', 'residue'
        create : bool = False
            Create the template if it does not exist.

        Returns
        -------
        int
            The id of the template.

        Raises
        ------
        KeyError
           If the template does not exist and 'create' is not requested.
        """
        self.cursor.execute(
            f'SELECT id FROM {self.table} WHERE "name" = ? AND "type" = ?',
            (name, type_)
        )
        row = self.cursor.fetchone()
        if row is None:
            if create:
                return self.create(name, type_=type_)
            else:
                raise KeyError(
                    f"There is no template '{name}' of type '{type_}'."
                )
        return row[0]

    def set_current_template(self, name, type_='general'):
        """Set the current template given the name and optionally type.

        Parameters
        ----------
        name : str
            The name of the template. The name/type pair must be unique.
        type_ : str = 'general'
            The type of template.

        Returns
        -------
        None
        """
        id = self.find(name, type_=type_)
        self._current_template = id

    def templates(self, name=None, type_=None):
        """Return an itereator over the given templates.

        Parameters
        ----------
        name : str
            The name of the template.
        type_ : str = 'general'
            The type of template, e.g. 'all', 'molecule', 'residue'

        Returns
        -------
        sqlite.cursor
            The iterator over rows.
        """
        if name is None:
            if type_ is None:
                return self.db.execute(f'SELECT * FROM {self.table}')
            else:
                return self.db.execute(
                    f'SELECT * FROM {self.table} WHERE "type" = ?', (type_,)
                )
        elif type_ is None:
            return self.db.execute(
                f'SELECT * FROM {self.table} WHERE "name" = ?', (name,)
            )
        else:
            return self.db.execute(
                f'SELECT * FROM {self.table} WHERE "name" = ? AND "type" = ?',
                (name, type_)
            )



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding atoms for the templates
"""

from itertools import zip_longest
import logging
from typing import Any, Dict, TypeVar

from molsystem.atoms import _Atoms as Atoms
from molsystem.column import _Column as Column
from molsystem.table import _Table as Table

System_tp = TypeVar("System_tp", "System", None)
Atoms_tp = TypeVar("Atoms_tp", "_Atoms", str, None)

logger = logging.getLogger(__name__)


def grouped(iterable, n):
    "s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,...s3n-1), ..."
    return zip_longest(*[iter(iterable)] * n)


class _Templateatoms(Atoms):
    """The Atoms class holds arrays of attributes describing atoms

    This is a bit complicated due to the separation of the actual atoms and the
    coordinates, which depend on the configuration. Also, the list of atoms can
    itself be time-dependent, and is controlled by the subset 'all'.

    Atoms can be added ('append') or removed ('delete').
    """

    def __init__(
        self,
        system: System_tp,
        atom_tablename='templateatom',
        coordinates_tablename='templatecoordinates',
    ) -> None:

        self._system = system
        self._atom_tablename = atom_tablename
        self._coordinates_tablename = coordinates_tablename

        self._atom_table = Table(system, self._atom_tablename)
        self._coordinates_table = Table(system, self._coordinates_tablename)
        self._templates = system['template']

    def __getitem__(self, key) -> Any:
        """Allow [] to access the data!"""
        if key in self._atom_table.attributes:
            sql = f'WHERE template = {self.current_template}'
            return Column(self._atom_table, key, where=sql)
        elif key in self._coordinates_table.attributes:
            where = (
                "WHERE templateatom in ("
                f"     SELECT id FROM {self._atom_tablename}"
                f"      WHERE template = {self.current_template}"
                ")"
            )
            return Column(self._coordinates_table, key, where=where)
        else:
            raise KeyError(f"'{key}' not in template atoms")

    @property
    def current_template(self):
        """The current template in use."""
        return self._templates.current_template

    @current_template.setter
    def current_template(self, value):
        self._templates.current_template = value

    @property
    def n_atoms(self) -> int:
        """The number of atoms *in the current* template."""
        self.cursor.execute(
            f"SELECT COUNT(*) FROM {self._atom_tablename} WHERE template = ?",
            (self.current_template,)
        )
        return self.cursor.fetchone()[0]

    @property
    def attributes(self) -> Dict[str, Any]:
        """The definitions of the attributes.
        Combine the attributes of the atom and coordinates tables to
        make it look like a single larger table.
        """

        result = self._atom_table.attributes

        for key, value in self._coordinates_table.attributes.items():
            if key != 'templateatom':  # ignore foreign key linking tables
                result[key] = value

        return result

    @property
    def coordinate_system(self):
        """The type of coordinates: 'fractional' or 'Cartesian'"""
        return 'Cartesian'

    @coordinate_system.setter
    def coordinate_system(self, value):
        raise RuntimeError('Templates can only use Cartesian coordinates.')

    def append(self, n=None, **kwargs: Dict[str, Any]) -> None:
        """Append one or more atoms

        The keys give the field for the data. If an existing field is not
        mentioned, then the default value is used, unless the default is None,
        in which case an error is thrown. It is an error if there is not a
        field corrresponding to a key.
        """

        # Need to handle the elements specially. Can give atomic numbers,
        # or symbols. By construction the references to elements are identical
        # to their atomic numbers.

        if 'symbol' in kwargs:
            symbols = kwargs.pop('symbol')
            kwargs['atno'] = self.to_atnos(symbols)

        # How many new rows there are
        n_rows, lengths = self._get_n_rows(**kwargs)

        if n is not None:
            if n_rows != 1 and n_rows != n:
                raise RuntimeError(
                    f"Requested number of template atoms ({n}) is not "
                    f"compatible with the length of the data ({n_rows})."
                )
            n_rows = n

        # Fill in the atom table
        data = {}
        for column in self._atom_table.attributes:
            if column != 'id' and column in kwargs:
                data[column] = kwargs.pop(column)
        if 'template' not in data:
            data['template'] = self.current_template

        ids = self._atom_table.append(n=n_rows, **data)

        # Now append to the coordinates table, but only if needed.
        data = {}
        for column in self._coordinates_table.attributes:
            if column != 'templateatom' and column in kwargs:
                data[column] = kwargs.pop(column)

        if len(data) > 0:
            data['templateatom'] = ids
            self._coordinates_table.append(n=n_rows, **data)

        return ids

    def atomic_numbers(self, template: int = None) -> [int]:
        """The atomic numbers of the atoms in a template.

        Parameters
        ----------
        template : int = None
            Which template, defaulting to the current template.

        Returns
        -------
        ids : [int]
            The ids of the atoms in the template.
        """
        if template is None:
            template = self.current_template
        return [
            x[0] for x in self.db.execute(
                f"SELECT atno FROM {self._atom_tablename} WHERE template = ?",
                (template,)
            )
        ]

    def atom_ids(self, template: int = None) -> [int]:
        """The ids of the atoms in a template.

        Parameters
        ----------
        template : int = None
            Which template, defaulting to the current template.

        Returns
        -------
        ids : [int]
            The ids of the atoms in the template.
        """
        if template is None:
            template = self.current_template
        return [
            x[0] for x in self.db.execute(
                f"SELECT id FROM {self._atom_tablename} WHERE template = ?",
                (template,)
            )
        ]

    def atoms(self, *args, template=None):
        """Get an iterator over atoms in the template.

        Parameters
        ----------
        args : str, int or float
            SQL restrictions for a WHERE statement, each argument being
            one word, e.g. "atno" "=" 5
        template : int = None
            Which template, defaulting to the current template.

        Returns
        -------
        SQLite3.Cursor :
            The cursor containing the result.
        """
        atom_tbl = self._atom_tablename
        coord_tbl = self._coordinates_tablename
        atom_columns = [*self._atom_table.attributes]
        coord_columns = [*self._coordinates_table.attributes]
        coord_columns.remove('templateatom')

        columns = [f'{atom_tbl}.{x}' for x in atom_columns]
        columns += [f'{coord_tbl}.{x}' for x in coord_columns]
        column_defs = ', '.join(columns)

        if template is None:
            template = self.current_template

        sql = (
            f'SELECT {column_defs} FROM {atom_tbl}, {coord_tbl}'
            f' WHERE {coord_tbl}.templateatom = {atom_tbl}.id'
            f'   AND {atom_tbl}.template = ?'
        )
        if len(args) == 0:
            return self.db.execute(sql, (template,))

        parameters = [template]
        for col, op, value in grouped(args, 3):
            if op == '==':
                op = '='
            sql += f' AND "{col}" {op} ?'
            parameters.append(value)

        return self.db.execute(sql, parameters)

    def remove(self, *args, template=None):
        """Remove atoms in the template.

        Parameters
        ----------
        args : str, int or float
            SQL restrictions for a WHERE statement, each argument being
            one word, e.g. "atno" "=" 5
        template : int = None
            Which template, defaulting to the current template.

        Returns
        -------
        None
        """
        if template is None:
            template = self.current_template

        sql = f'DELETE FROM {self.table} WHERE template = ?'

        parameters = [template]
        for col, op, value in grouped(args, 3):
            if op == '==':
                op = '='
            sql += f' AND "{col}" {op} ?'
            parameters.append(value)

        self.db.execute(sql, parameters)
        self.db.commit()



# -*- coding: utf-8 -*-

"""Functions for handling CIF files"""

import logging

logger = logging.getLogger(__name__)

bond_order = {1: 'SING', 2: 'DOUB', 3: 'TRIP'}


class CIFMixin:
    """A mixin for handling CIF files."""

    def to_cif_text(self, configuration=None):
        """Create the text of a CIF file from the confguration.

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.

        Returns
        -------
        text : str
            The text of the file.
        """

        lines = []

        atoms = self.atoms

        if configuration is None:
            configuration = self.current_configuration

        # Get the chemical formula
        formula, empirical_formula, Z = self.formula(configuration)
        formula = ' '.join(formula)
        empirical_formula = ''.join(empirical_formula)

        # And created the file, line-by-line
        lines = []
        lines.append('# Generated by MolSSI SEAMM')
        lines.append(f"data_{empirical_formula}")

        # Cell information
        if self.periodicity == 3:
            cell = self['cell'].cell(configuration)
            a, b, c, alpha, beta, gamma = cell.parameters
            volume = cell.volume
            lines.append("_symmetry_space_group_name_H-M   'P 1'")
            lines.append(f'_cell_length_a   {a}')
            lines.append(f'_cell_length_b   {b}')
            lines.append(f'_cell_length_c   {c}')
            lines.append(f'_cell_angle_alpha   {alpha}')
            lines.append(f'_cell_angle_beta    {beta}')
            lines.append(f'_cell_angle_gamma   {gamma}')
            lines.append('_symmetry_Int_Tables_number   1')
            lines.append(f'_cell_volume   {volume}')
            lines.append(f'_cell_formula_units_Z   {Z}')
            lines.append('loop_')
            lines.append(' _symmetry_equiv_pos_site_id')
            lines.append(' _symmetry_equiv_pos_as_xyz')
            lines.append("  1  'x, y, z'")

        lines.append(f'_chemical_formula_structural   {empirical_formula}')
        lines.append(f"_chemical_formula_sum   '{formula}'")

        # The atoms
        lines.append('loop_')
        lines.append(' _atom_site_type_symbol')
        lines.append(' _atom_site_label')
        lines.append(' _atom_site_symmetry_multiplicity')
        lines.append(' _atom_site_fract_x')
        lines.append(' _atom_site_fract_y')
        lines.append(' _atom_site_fract_z')
        lines.append(' _atom_site_occupancy')

        # Need unique names
        if 'names' in atoms:
            original_names = atoms.get_column('names', configuration)
        else:
            original_names = atoms.symbols(configuration)

        names = []
        tmp = {}
        for name in original_names:
            if name in tmp:
                tmp[name] += 1
                names.append(name + str(tmp[name]))
            else:
                tmp[name] = 1
                names.append(name)

        UVW = atoms.coordinates(
            configuration=configuration, fractionals=True, in_cell='molecule'
        )

        symbols = atoms.symbols(configuration)
        for element, name, uvw in zip(symbols, names, UVW):
            u, v, w = uvw
            lines.append(f'{element} {name}  1  {u:.3f} {v:.3f} {w:.3f}  1')

        # And that is it!
        return '\n'.join(lines)

    def to_mmcif_text(self, configuration=None):
        """Create the text of a mmCIF file from the confguration.

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.

        Returns
        -------
        text : str
            The text of the file.
        """

        lines = []

        atoms = self.atoms
        bonds = self.bonds

        if configuration is None:
            configuration = self.current_configuration

        # Get the chemical formula
        formula, empirical_formula, Z = self.formula(configuration)
        formula = ' '.join(formula)
        empirical_formula = ''.join(empirical_formula)

        # And created the file, line-by-line
        lines = []
        lines.append('# Generated by MolSSI SEAMM')
        lines.append(f"data_{empirical_formula}")
        lines.append(f"_chem_comp.name '{formula}'")
        lines.append(f"_chem_comp.id '{empirical_formula}'")
        lines.append(f"_chem_comp.formula   '{formula}'")

        # Cell information
        if self.periodicity == 3:
            cell = self['cell'].cell(configuration)
            a, b, c, alpha, beta, gamma = cell.parameters
            volume = cell.volume
            lines.append("_symmetry_space_group_name_H-M   'P 1'")
            lines.append(f'_cell_length_a   {a}')
            lines.append(f'_cell_length_b   {b}')
            lines.append(f'_cell_length_c   {c}')
            lines.append(f'_cell_angle_alpha   {alpha}')
            lines.append(f'_cell_angle_beta    {beta}')
            lines.append(f'_cell_angle_gamma   {gamma}')
            lines.append('_symmetry_Int_Tables_number   1')
            lines.append(f'_cell_volume   {volume}')
            lines.append(f'_cell_formula_units_Z   {Z}')
            lines.append('loop_')
            lines.append(' _symmetry_equiv_pos_site_id')
            lines.append(' _symmetry_equiv_pos_as_xyz')
            lines.append("  1  'x, y, z'")

        lines.append(f'_chemical_formula_structural   {empirical_formula}')
        lines.append(f"_chemical_formula_sum   '{formula}'")

        # The atoms
        lines.append('loop_')
        lines.append(' _atom_site_type_symbol')
        lines.append(' _atom_site_label')
        lines.append(' _atom_site_symmetry_multiplicity')
        lines.append(' _atom_site_fract_x')
        lines.append(' _atom_site_fract_y')
        lines.append(' _atom_site_fract_z')
        lines.append(' _atom_site_occupancy')

        # Need unique names
        if 'names' in atoms:
            original_names = atoms.get_column('names', configuration)
        else:
            original_names = atoms.symbols(configuration)

        names = []
        tmp = {}
        for name in original_names:
            if name in tmp:
                tmp[name] += 1
                names.append(name + str(tmp[name]))
            else:
                tmp[name] = 1
                names.append(name)

        XYZ = atoms.coordinates(
            configuration=configuration, in_cell='molecule'
        )
        XYZa = atoms.coordinates(configuration=configuration)

        symbols = atoms.symbols(configuration)
        for element, name, xyza, xyz in zip(symbols, names, XYZa, XYZ):
            xa, ya, za = xyza
            x, y, z = xyz
            lines.append(
                f'MOL1 {name} {element} {xa:.3f} {ya:.3f} {za:.3f} '
                f'{x:.3f} {y:.3f} {z:.3f} HET 1'
            )

        # The bonds
        lines.append('#')
        lines.append('loop_')
        lines.append(' _chem_comp_bond.comp_id')
        lines.append(' _chem_comp_bond.atom_id_1')
        lines.append(' _chem_comp_bond.atom_id_2')
        lines.append(' _chem_comp_bond.value_order')
        for row in bonds.bonds(configuration=configuration):
            i = row['i']
            j = row['j']
            order = bond_order[row['bondorder']]
            nm1 = names[i - 1]
            nm2 = names[j - 1]
            lines.append(f'MOL1 {nm1} {nm2} {order}')

        # And that is it!
        return '\n'.join(lines)



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding a system
"""

import collections.abc
from collections import Counter
from functools import reduce
import logging
import math
import sqlite3
from typing import Any, Dict

from molsystem.elemental_data import element_data
from molsystem.table import _Table as Table
from molsystem.atoms import _Atoms as Atoms
from molsystem.subset import _Subsets as Subsets
from molsystem.template import _Template as Template
from molsystem.templateatoms import _Templateatoms as Templateatoms
from molsystem.templatebonds import _Templatebonds as Templatebonds
from molsystem.bonds import _Bonds as Bonds
from molsystem.cell_parameters import _CellParameters as CellParameters

from molsystem.cif import CIFMixin
from molsystem.molfile import MolFileMixin
from molsystem.pdb import PDBMixin
from molsystem.smiles import SMILESMixin
from molsystem.topology import TopologyMixin

logger = logging.getLogger(__name__)


class _System(
    PDBMixin, MolFileMixin, CIFMixin, SMILESMixin, TopologyMixin,
    collections.abc.MutableMapping
):
    """A single system -- molecule, crystal, etc. -- in SEAMM.

    Based on a SQLite database, this class provides a general
    datastructure for storing a 'system', which is an atom, molecule,
    crystal or a collection of any number of such components all held in
    a single system, in analogy to a thermodynamic system, which is what
    an experiment is perfromed on.

    The system can hold multiple different configurations of the atoms,
    such as found in an MD trajectory or conformer search. The number of
    atoms may change across configurations as may the bonding, so the
    system is capabale of handling a trajectory for e.g. a grand
    canonical simulation or simulating desposition on a surface. The
    system also can contain bonds as are used in classical valence
    forcefields, and can also cope with changing bonds in e.g. reactive
    forcefields.

    The system class also supports the concept of template molecules or
    subsets, which can be used to define molecules, residues, etc. which
    are then repeated in the full structure. Templates make the process
    of building complex, repeated structures simpler and more efficient,
    and also help track the individual components. There is a lighter
    variation of templates called subsets which do not define a
    substructure, but can label and track such substructures. Atoms may
    belong to more than one subset, or to none other than the special
    subset 'all' which contains all the atoms in a given configuration.

    The key tables are:

    system -- a container for all the components, plus storage for the
        system-wide parameters and options.

    configuration -- a list of configurations in e.g. a trajectory.
    configuration_subset -- connnects configurations with subsets.

    cell -- the information about the periodicity of a configuration, if
        needed.

    symmetry -- the point or space group symmetry of the configuration.
    symop -- the symmetry operations for the group.

    template -- a list of all templates.
    templateatom -- holds the atoms in a template, if present.
    templatebond -- holds the bonds between the template atoms, if present.

    subset -- an instantiation of a template, connected with one or more
        configurations of the system.
    subset_atom -- connects the subset to the atoms, and optionally connects
        the atom to a template atom.

    atom -- the nonvarying part of the description of the atoms
    coordinates -- the varying part of the description of atoms

    element -- basic information about the elements. The atomic number
        (atno) is used for foreign keys in other tables.

    This system class provides an anchor for the other component and
    handles system-wide information such as the periodicity and the
    coordinate system used.

    It also tracks the current configuration and template, which are
    conveniences to avoid having to specify the configuration or
    template for every access to the data.

    There is one special, required subset 'all' that contains all of the
    atoms in each configuration of the system. If there are differing
    numbers or types of atoms in the system, or if the bonding changes,
    there will be multiple versions of the subset, each corresponding to
    a different set of atoms or bonds.

    One or more configurations are connected with each 'all' subset,
    which is how the atoms and bonding are connected to the
    configurations.
    """

    def __init__(self, parent, nickname=None, **kwargs):
        self._parent = parent
        self._nickname = nickname
        self._attached = []
        self._id = 1  # The id of the system. Currently there is only 1
        self._current_configuration = None  # The current configuration
        self._configurations = {}  # template and subset all for configs
        self._checkpoints = []
        self._filename = None
        self._db = None
        self._cursor = None
        self._items = {}
        self._symbol_to_atno = {}
        self._atno_to_symbol = {}

        if 'filename' in kwargs:
            self.filename = kwargs.pop('filename')
        else:
            self.filename = 'seamm.db'

    def __enter__(self) -> Any:
        self.db.commit()

        backup = self.parent.copy_system(self, temporary=True)
        self._checkpoints.append(backup)
        return self

    def __exit__(self, etype, value, traceback) -> None:
        backup = self._checkpoints.pop()

        if etype is None:
            self.db.commit()

            # Log the changes
            diffs = self.diff(backup)
            if len(diffs) > 0:
                self.version = self.version + 1

            # Not sure why this commit is needed...
            self.db.commit()
        else:
            self.parent.overwrite(self, backup)

        # and delete the copy
        del self.parent[backup.nickname]

    def __getitem__(self, key):
        """Allow [] access to the dictionary!

        Because some of the items, such as template contain state,
        we need to ensure that the same object is used everywhere. Hence
        the self._items array to sotre the instances.

        Parameters
        ----------
        key : str
            The table name or name of a multi-table item like atoms.

        Returns
        -------
        table : Table or Atom, Template, Templateatoms, ...
        """
        if key == 'atom' or key == 'atoms':
            if 'atom' not in self._items:
                self._items['atom'] = Atoms(self)
            return self._items['atom']
        elif key == 'template' or key == 'templates':
            if 'template' not in self._items:
                self._items['template'] = Template(self)
            return self._items['template']
        elif key == 'templateatom' or key == 'templateatoms':
            if 'templateatom' not in self._items:
                self._items['templateatom'] = Templateatoms(self)
            return self._items['templateatom']
        elif key == 'templatebond' or key == 'templatebonds':
            if 'templatebond' not in self._items:
                self._items['templatebond'] = Templatebonds(self)
            return self._items['templatebond']
        elif key == 'bond' or key == 'bonds':
            if 'bond' not in self._items:
                self._items['bond'] = Bonds(self)
            return self._items['bond']
        elif key == 'subset' or key == 'subsets':
            if 'subset' not in self._items:
                self._items['subset'] = Subsets(self)
            return self._items['subset']
        elif key == 'cell':
            if 'cell' not in self._items:
                self._items['cell'] = CellParameters(self)
            return self._items['cell']
        else:
            if key not in self._items:
                self._items[key] = Table(self, key)
            return self._items[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        raise NotImplementedError(f"Table '{key}' cannot be created yet")

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self:
            self.cursor.execute(f"DROP TABLE '{key}'")

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self.list())

    def __len__(self):
        """The len() command"""
        self.cursor.execute(
            "SELECT COUNT(*)"
            "  FROM sqlite_master"
            " WHERE type = 'table'"
        )
        return self.cursor.fetchone()[0]

    # def __repr__(self):
    #     """The string representation of this object"""
    #     raise NotImplementedError()

    # def __str__(self):
    #     """The pretty string representation of this object"""
    #     raise NotImplementedError()

    def __contains__(self, item):
        """Return a boolean indicating if a key exists."""
        # Normal the tablename is used as an identifier, so is quoted with ".
        # Here we need it as a string literal so strip any quotes from it.

        tmp_item = item.strip('"')
        self.cursor.execute(
            "SELECT COUNT(*)"
            "  FROM sqlite_master"
            f" WHERE type = 'table' and name = '{tmp_item}'"
        )
        return self.cursor.fetchone()[0] == 1

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        if self is other:
            return True

        tables = set(self.list())
        other_tables = set(other.list())

        added = tables - other_tables
        removed = other_tables - tables
        in_common = tables & other_tables

        if len(added) > 0:
            return False
        if len(removed) > 0:
            return False

        # Need the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        name = self.name
        other_name = other.name
        detach = False
        if name != other_name and not self.is_attached(other_name):
            # Attach the other system in order to do comparisons.
            self.attach(other)
            detach = True

        # Check the tables in both systems
        result = True
        for table in in_common:
            if Table(self, table) != Table(other, table):
                result = False
                break

        # Detach the other database if needed
        if detach:
            self.detach(other)

        return result

    @property
    def atoms(self):
        """The atoms, which are held as a dictionary of arrays"""
        return self['atom']

    @property
    def bonds(self):
        """The bonds, which are held as a dictionary of arrays"""
        return self['bond']

    @property
    def cell(self):
        """The periodic cell."""
        return self['cell']

    @property
    def configurations(self):
        """The dictionary of configurations."""
        return self._configurations

    @property
    def coordinate_system(self):
        """The coordinates system used, 'fractional' or 'Cartesian'"""
        self.cursor.execute(
            "SELECT coordinatesystem FROM system WHERE id = ?", (self._id,)
        )
        return self.cursor.fetchone()[0]

    @coordinate_system.setter
    def coordinate_system(self, value):
        if value.lower()[0] == 'f':
            self.cursor.execute(
                "UPDATE system SET coordinatesystem = 'fractional'"
                " WHERE id = ?", (self._id,)
            )
        else:
            self.cursor.execute(
                "UPDATE system SET coordinatesystem = 'Cartesian'"
                " WHERE id = ?", (self._id,)
            )

    @property
    def current_configuration(self):
        """The current configuration to work with."""
        return self._current_configuration

    @current_configuration.setter
    def current_configuration(self, value):
        if value not in self.configurations:
            raise KeyError(f"configuration '{value}' doe not exist.")
        self._current_configuration = value

    @property
    def cursor(self):
        return self._cursor

    @property
    def db(self):
        return self._db

    @property
    def filename(self):
        """The name of the file (or URI) for the database."""
        return self._filename

    @filename.setter
    def filename(self, value):
        if value != self._filename:
            if self._db is not None:
                self.cursor.close()
                self._db.commit()
                self._db.close()
                self._db = None
                self._cursor = None
            self._filename = value
            if self._filename is not None:
                self._db = sqlite3.connect(self._filename)
                self._db.row_factory = sqlite3.Row
                self._db.execute('PRAGMA foreign_keys = ON')
                self._cursor = self._db.cursor()
                self._initialize()

    @property
    def name(self):
        """Return the name of this system."""
        self.cursor.execute(
            "SELECT name FROM system WHERE id = ?", (self._id,)
        )
        result = self.cursor.fetchone()
        if result is None:
            return None
        else:
            return result[0]

    @name.setter
    def name(self, value):
        self.cursor.execute(
            "UPDATE system SET name = ? WHERE id = ?", (value, self._id)
        )

    @property
    def nickname(self):
        """The name used in Systems for this system"""
        return self._nickname

    @property
    def n_configurations(self):
        """The number of configurations of the system."""
        return self['configuration'].n_rows

    @property
    def periodicity(self):
        """The periodicity of the system, 0, 1, 2 or 3"""
        self.cursor.execute(
            "SELECT periodicity FROM system WHERE id=?", (self._id,)
        )
        return self.cursor.fetchone()[0]

    @property
    def parent(self):
        """The parent of this, i.e. a Systems object."""
        return self._parent

    @periodicity.setter
    def periodicity(self, value):
        if value < 0 or value > 3:
            raise ValueError('The periodicity must be between 0 and 3.')
        self.cursor.execute(
            "UPDATE system SET periodicity = ? WHERE id = ?",
            (value, self._id)
        )

    @property
    def subsets(self):
        """The subsets"""
        return self['subset']

    @property
    def templates(self):
        """The templates"""
        return self['template']

    @property
    def templateatoms(self):
        """The template atoms"""
        return self['templateatom']

    @property
    def templatebonds(self):
        """The template bonds"""
        return self['templatebond']

    @property
    def version(self):
        """The version of the system, incrementing from 0"""
        self.cursor.execute("SELECT version FROM system")
        return int(self.cursor.fetchone()[0])

    @version.setter
    def version(self, value):
        self.cursor.execute("UPDATE system SET version = ?", (str(value),))
        self.db.commit()

    def add_configuration(
        self,
        system=1,
        name=None,
        symmetry=None,
        cell=None,
        changed_atoms=False,
        changed_bonds=False
    ):
        """Add a new configuration to the system.

        Parameters
        ----------
        system : int = 1
            The system for which this is a configuration. (optional)
        name : str = None
            A textual name for the configuration (optional)
        symmetry : int or str = None
            The id or name of the point or space group (optional)
        cell : Cell or 6-vector = None
            The cell parameters, default to last ones.
        changed_atoms : bool = False
            Whether the atoms have changed in number or identity from the
            previous configuration.
        changed_bonds : bool = False
            Whether the bonding has changed from the previous configuration.

        Returns
        -------
        configuration : int
            The id of the new configuration.

        What needs to be done depends on whether the atoms or bonds are
        changing:

        Case 1: The bonds are changing.
            * create a new template
            * create a subset for the new template
            * if the atoms aren't changing copy the previous connections in
              subset_atom to this instance.
            This routine does not create any templateatoms or bonds.

        Case 2: The atoms are changing
            * create new subset using the previous template
            This routine does not connect the atoms to the new 'all' subset,
            nor does it add any templateatoms or bonds to the new template.

        Case 3: Neither the atoms nor bonding is changing
            * Connect the previous 'all' subset to the configuration in the
              configuration_subset table.
        """
        # Work out the subset and template for 'all'
        if len(self._configurations) == 0:
            changed_bonds = True
            changed_atoms = True
            tid = self['template'].append(name='all', type='all')[0]
            sid = self['subset'].create(template=tid)
        else:
            last_configuration = max(self._configurations)
            last_sid, last_tid = self._configurations[last_configuration]
            sid = last_sid
            tid = last_tid

            if changed_bonds:
                # Case 1
                # If the bonding changed, need a new template and new subset
                tid = self['template'].append(name='all', type='all')
                sid = self['subset'].append(template=tid)
                if not changed_atoms:
                    atom_ids = self['atom'].atoms(
                        configuration=last_configuration
                    )
                    self['subset_atom'].append(subset=sid, atom=atom_ids)
            elif changed_atoms:
                # Case 2
                sid = self['subset'].append(template=tid)[0]
            else:
                # Case 3
                pass

        cid = self['configuration'].append(
            system=system, name=name, symmetry=symmetry
        )[0]

        self['configuration_subset'].append(configuration=cid, subset=sid)
        self._configurations[cid] = (sid, tid)

        return cid

    def all_subset(self, configuration=None):
        if configuration is None:
            configuration = self.current_configuration
        return self._configurations[configuration][0]

    def all_template(self, configuration=None):
        if configuration is None:
            configuration = self.current_configuration
        return self._configurations[configuration][1]

    def attach(self, other):
        """Attach another system to this one's database."""
        if self.is_attached(other.nickname):
            return
        self.db.execute(
            f"ATTACH DATABASE '{other.filename}' AS '{other.nickname}'"
        )
        self._attached.append(other.nickname)

    def is_attached(self, name):
        """Return whether another system is attached to this one."""
        return name in self._attached

    def detach(self, other):
        """Detach an attached system."""
        if self.is_attached(other.name):
            self.cursor.execute(f'DETACH DATABASE "{other.name}"')
            self._attached.remove(other.name)

    def clear(self, configuration=None) -> int:
        """Remove everything from the configuration

        Parameters
        ----------
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration. Not used if the subset is given.

        Returns
        -------
        int
            Number of atoms
        """
        # Delete the atoms
        self.atoms.remove(configuration=configuration)

        # Delete the template atoms.
        self.templateatoms.remove(template=self.all_template(configuration))

    def create_table(self, name, cls=Table, other=None):
        """Create a new table with the given name.

        Parameters
        ----------
        name : str
            The name of the new table.

        cls : Table subclass
            The class of the new table, defaults to Table

        Returns
        -------
        table : class Table
            The new table
        """
        if name in self:
            raise KeyError(f"'{name}' already exists in the system.")

        self._items[name] = cls(self, name, other)
        return self._items[name]

    def diff(self, other):
        """Differences between this system and another."""
        result = {}

        if self is other:
            return result

        tables = set(self.list())
        other_tables = set(other.list())

        added = tables - other_tables
        removed = other_tables - tables
        in_common = tables & other_tables

        if len(added) > 0:
            result['tables added'] = list(added)
        if len(removed) > 0:
            result['tables removed'] = list(removed)

        # Need the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        name = self.nickname
        other_name = other.nickname
        detach = False
        if name != other_name and not self.is_attached(other_name):
            # Attach the other system in order to do comparisons.
            self.attach(other)
            detach = True

        # Check the tables in both systems
        for table in in_common:
            tmp = Table(self, table).diff(Table(other, table))
            if len(tmp) > 0:
                result[f"table '{table}' diffs"] = tmp

        # Detach the other database if needed
        if detach:
            self.detach(other)

        return result

    def formula(self, configuration=None):
        """Return the chemical formula of the configuration.

        Returns a tuple with the formula, empirical formula and number of
        formula units (Z).

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.

        Returns
        -------
        formulas : (str, str, int)
            The chemical formula, empirical formula and Z.
        """
        counts = Counter(self.atoms.symbols(configuration))

        # Order the elements ... Merck CH then alphabetical,
        # or if no C or H, then just alphabetically
        formula_list = []
        if 'C' in counts and 'H' in counts:
            formula_list.append(('C', counts.pop('C')))
            formula_list.append(('H', counts.pop('H')))

        for element in sorted(counts.keys()):
            formula_list.append((element, counts[element]))

        counts = []
        for _, count in formula_list:
            counts.append(count)

        formula = []
        for element, count in formula_list:
            if count > 1:
                formula.append(f'{element}{count}')
            else:
                formula.append(element)

        # And the empirical formula
        Z = reduce(math.gcd, counts)
        empirical_formula_list = []
        for element, count in formula_list:
            empirical_formula_list.append((element, int(count / Z)))

        empirical_formula = []
        for element, count in empirical_formula_list:
            if count > 1:
                empirical_formula.append(f'{element}{count}')
            else:
                empirical_formula.append(element)

        return formula, empirical_formula, Z

    def list(self):
        """Return a list of all the tables in the system."""
        result = []
        for row in self.db.execute(
            "SELECT name"
            "  FROM sqlite_master"
            " WHERE type = 'table'"
        ):
            result.append(row['name'])
        return result

    def n_atoms(self, subset=None, configuration=None) -> int:
        """The number of atoms in a subset or configuration

        Parameters
        ----------
        subset : int = None
            Get the atoms for the subset. Defaults to the 'all/all' subset
            for the configuration given.
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration. Not used if the subset is given.

        Returns
        -------
        int
            Number of atoms
        """
        return self.atoms.n_atoms(subset=subset, configuration=configuration)

    def n_bonds(self, subset: int = None, configuration: int = None) -> int:
        """The number of bonds.

        Parameters
        ----------
        bonds : int = None
            Get the bonds for the subset. Defaults to the 'all/all' subset
            for the configuration given.
        configuration : int = None
            The configuration of interest. Defaults to the current
            configuration. Not used if the subset is given.

        Returns
        -------
        int
            Number of bonds
        """
        return self.bonds.n_bonds(subset=subset, configuration=configuration)

    def to_atnos(self, symbols):
        """Convert element symbols to atomic numbers.

        Parameters
        ----------
        symbols : [str]
            The atomic symbols

        Returns
        -------
        atnos : [int]
            The corresponding atomic numbers (1..118)
        """
        return [self._symbol_to_atno[x] for x in symbols]

    def to_symbols(self, atnos):
        """Convert atomic numbers to element symbols.

        Parameters
        ----------
        atnos : [int]
            The atomic numbers (1..118)

        Returns
        -------
        symboles : [str]
            The corresponding atomic symbols
        """
        return [self._atno_to_symbol[x] for x in atnos]

    def _initialize(self):
        """Initialize the SQLite database."""
        if 'element' not in self:
            self._initialize_elements()

        if 'symmetry' not in self:
            self._initialize_symmetry()

        if 'cell' not in self:
            self._initialize_cell()

        if 'system' not in self:
            self._initialize_system()
            self.name = self._nickname
        self._id = 1

        if 'configuration' not in self:
            self._initialize_configurations()
        else:
            # Get all the configurations from the database
            for row in self.db.execute(
                "SELECT configuration, subset, template FROM "
                "       configuration_subset, subset, template"
                " WHERE template.name = 'all' AND template.type = 'all'"
                "   AND template.id = template AND subset.id = subset"
            ):
                config = row['configuration']
                self._configurations[config] = (row['subset'], row['template'])
        if 'atom' not in self:
            self._initialize_atoms()

        if 'subset' not in self:
            self._initialize_subsets()

        # If needed, set up the first configuration, and the 'all' subset
        if self.n_configurations == 0:
            self.current_configuration = self.add_configuration()

    def _initialize_system(self):
        """Set up the table for the system."""
        table = self['system']
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute(
            'name', coltype='str', notnull=True, default='default'
        )
        table.add_attribute('version', coltype='int', notnull=True, default=0)
        table.add_attribute(
            'periodicity', coltype='int', notnull=True, default=0
        )
        table.add_attribute(
            'coordinatesystem',
            coltype='str',
            notnull=True,
            default='Cartesian'
        )

        table.append(
            id=1,
            name='default',
            version=0,
            periodicity=0,
            coordinatesystem='Cartesian'
        )
        self.db.commit()

    def _initialize_atoms(self):
        """Set up the tables for atoms."""
        table = Table(self, 'atom')
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute('atno', coltype='int', references='element')

        table = self['coordinates']
        table.add_attribute(
            'configuration', coltype='int', references='configuration'
        )
        table.add_attribute('atom', coltype='int', references='atom')
        table.add_attribute('x', coltype='float')
        table.add_attribute('y', coltype='float')
        table.add_attribute('z', coltype='float')

    def _initialize_cell(self):
        """Set up the tables for the cell."""
        table = self['cell']
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute('a', coltype='float', default=10.0)
        table.add_attribute('b', coltype='float', default=10.0)
        table.add_attribute('c', coltype='float', default=10.0)
        table.add_attribute('alpha', coltype='float', default=90.0)
        table.add_attribute('beta', coltype='float', default=90.0)
        table.add_attribute('gamma', coltype='float', default=90.0)

    def _initialize_elements(self):
        """Set up the table of elements."""
        table = self['element']
        table.add_attribute('atno', coltype='int', pk=True)
        table.add_attribute('symbol', coltype='str', index='unique')
        table.add_attribute('mass', coltype='float')

        for symbol, data in element_data.items():
            table.append(
                symbol=symbol,
                atno=data['atomic number'],
                mass=data['atomic weight']
            )
            self._symbol_to_atno[symbol] = data['atomic number']
            self._atno_to_symbol[data['atomic number']] = symbol
        self.db.commit()

    def _initialize_configurations(self):
        """Set up the table of configurations."""
        table = self['configuration']
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute('system', coltype='int', references='system')
        table.add_attribute('name', coltype='str')
        table.add_attribute('symmetry', coltype='int', references='symmetry')
        table.add_attribute('cell', coltype='int', references='cell')
        self.db.commit()

    def _initialize_subsets(self):
        """Set up the tables for handling subsets.
        """
        # The definition of the subsets -- templates
        table = self['template']
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute('name', coltype='str')
        table.add_attribute('type', coltype='str', default='general')
        self.db.execute(
            "CREATE UNIQUE INDEX 'idx_template_name_type'"
            '    ON template ("name", "type")'
        )

        # The subsets
        table = self['subset']
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute('template', coltype='int', references='template')

        # The template atoms (if any) for the subset
        table = Table(self, 'templateatom')
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute('template', coltype='int', references='template')
        table.add_attribute('name', coltype='str')
        table.add_attribute('atno', coltype='int', references='element')

        # The template coordinates (if any) for the subset
        table = Table(self, 'templatecoordinates')
        table.add_attribute(
            'templateatom', coltype='int', references='templateatom'
        )
        table.add_attribute('x', coltype='float')
        table.add_attribute('y', coltype='float')
        table.add_attribute('z', coltype='float')

        # And bonding for the template
        table = self['templatebond']
        table.add_attribute('i', coltype='int', references='templateatom')
        table.add_attribute('j', coltype='int', references='templateatom')
        table.add_attribute('bondorder', coltype='int', default=1)

        # The connection from the subsets to the configurations
        table = self['configuration_subset']
        table.add_attribute(
            'configuration', coltype='int', references='configuration'
        )
        table.add_attribute('subset', coltype='int', references='subset')

        # The connection between subsets and the atoms in the system
        table = self['subset_atom']
        table.add_attribute('atom', coltype='int', references='atom')
        table.add_attribute('subset', coltype='int', references='subset')
        table.add_attribute(
            'templateatom', coltype='int', references='templateatom'
        )

    def _initialize_symmetry(self):
        """Set up the tables for symmetry."""
        table = self['symmetry']
        table.add_attribute('id', coltype='int', pk=True)
        table.add_attribute('group', coltype='str')

        table = self['symmetryoperation']
        table.add_attribute('symmetry', coltype='int', references='symmetry')
        table.add_attribute('symop', coltype='str')

    def attributes(self, tablename: str) -> Dict[str, Any]:
        """The attributes -- columns -- of a given table.

        Parameters
        ----------
        tablename : str
            The name of the table, optionally including the schema followed by
            a dot.

        Returns
        -------
        attributes : Dict[str, Any]
            A dictionary of dictionaries for the attributes and their
            descriptors
        """
        if '.' in tablename:
            schema, tablename = tablename.split('.')
            sql = f"PRAGMA {schema}.table_info('{tablename}')"
        else:
            sql = f"PRAGMA table_info('{tablename}')"

        result = {}
        for line in self.db.execute(sql):
            result[line['name']] = {
                'type': line['type'],
                'notnull': bool(line['notnull']),
                'default': line['dflt_value'],
                'primary key': bool(line['pk'])
            }
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the topology mixin of the system class."""

import pprint

import pytest  # noqa: F401


def test_bonded_neighbors(AceticAcid):
    """Test the generation of bonded neighbors."""
    result = {
        1: [2, 3, 4, 5],
        2: [1],
        3: [1],
        4: [1],
        5: [1, 6, 7],
        6: [5],
        7: [5, 8],
        8: [7]
    }

    neighbors = AceticAcid.bonded_neighbors()
    if neighbors != result:
        pprint.pprint(neighbors)
    assert neighbors == result


def test_molecules(AceticAcid):
    """Test the finding molecules in the system."""
    result = [[1, 2, 3, 4, 5, 6, 7, 8]]

    molecules = AceticAcid.find_molecules()
    if molecules != result:
        pprint.pprint(molecules)
    assert molecules == result


def test_multiple_molecules(CH3COOH_3H2O):
    """Test the finding molecules in the system."""
    result = [
        [1, 2, 3, 4, 5, 6, 7, 8], [9, 10, 11], [12, 13, 14], [15, 16, 17]
    ]

    molecules = CH3COOH_3H2O.find_molecules()
    if molecules != result:
        pprint.pprint(molecules)
    assert molecules == result


def test_molecule_subsets(CH3COOH_3H2O):
    """Test making subsets for the molecules."""
    result = [2, 3, 4, 5]

    system = CH3COOH_3H2O
    sids = system.create_molecule_subsets()
    if sids != result:
        pprint.pprint(sids)
    assert sids == result

    sid = sids[1]
    assert system.atoms.n_atoms(subset=sid) == 3
    assert system.bonds.n_bonds(subset=sid) == 0  # No bonds defined!


def test_molecule_templates(disordered):
    """Test making templates for the molecules."""
    result_tids = [2, 2]
    result_sids = {2: [2, 3]}

    system = disordered
    tids, sids = system.create_molecule_templates()
    if tids != result_tids:
        print('tids')
        pprint.pprint(tids)
    if sids != result_sids:
        print('sids')
        pprint.pprint(sids)
    assert tids == result_tids
    assert sids == result_sids

    # And that the atoms are indeed order correctly
    tid = tids[0]
    sid1, sid2 = sids[tid]
    atnos1 = system.atoms.atomic_numbers(subset=sid1)
    atnos2 = system.atoms.atomic_numbers(subset=sid2, template_order=True)
    if atnos1 != atnos2:
        print(f'atnos1 = {atnos1}')
        print(f'atnos2 = {atnos2}')
    assert atnos1 == atnos2

    coords1 = system.atoms.coordinates(subset=sid1)
    coords2 = system.atoms.coordinates(subset=sid2, template_order=True)

    # At the moment the methyl hydrogens are not properly ordered -- they are
    # equivalent -- so remove them to test.
    c1 = list(coords1)
    del c1[1:4]
    c2 = list(coords2)
    del c2[1:4]

    if c1 != c2:
        print('coords1')
        pprint.pprint(coords1)
        print('coords2')
        pprint.pprint(coords2)
    assert c1 == c2



# -*- coding: utf-8 -*-

"""Functions for handling SMILES"""

import logging

from openbabel import openbabel

logger = logging.getLogger(__name__)


class SMILESMixin:
    """A mixin for handling SMILES."""

    def to_smiles(
        self,
        canonical=False,
        configuration=None,
        name=False,
        hydrogens=False
    ):
        """Create the SMILES string from the system.

        Parameters
        ----------
        canonical : bool = False
            Whether to create canonical SMILES
        configuration : int = None
            The configuration to use, defaults to the current configuration.
        name : bool = False
            Whether to return the name of the system
        hydrogens : bool = False
            Whether to keep H's in the SMILES string.

        Returns
        -------
        str
            The SMILES string, or (SMILES, name) if the rname is requested
        """
        logger.info('to_smiles')

        molfile = self.to_molfile_text()
        logger.info('...molfile = ')
        logger.info(molfile)
        logger.info('...end molfile')

        obConversion = openbabel.OBConversion()
        if canonical:
            obConversion.SetInAndOutFormats("mdl", "can")
        else:
            obConversion.SetInAndOutFormats("mdl", "smi")
        mol = openbabel.OBMol()
        obConversion.ReadString(mol, molfile)

        logger.info(f'MolFile has {mol.NumAtoms()} atoms')

        if not name:
            obConversion.AddOption('n')
        if hydrogens:
            obConversion.AddOption('h')
        smiles = obConversion.WriteString(mol)

        logger.info('')
        logger.info('')
        logger.info('')
        logger.info('')
        logger.info(f"smiles = '{smiles}'")

        if name:
            return smiles.strip().split('\t')
        else:
            return smiles.strip()

    def from_smiles(self, smiles, name=None, configuration=None):
        """Create the system from a SMILES string.

        Parameters
        ----------
        smiles : str
            The SMILES string
        name : str = None
            The name of the molecule
        configuration : int = None
            The configuration to use, defaults to the current configuration.

        Returns
        -------
        None
        """

        obConversion = openbabel.OBConversion()
        obConversion.SetInAndOutFormats("smi", "mdl")
        obConversion.AddOption('3')
        mol = openbabel.OBMol()
        obConversion.ReadString(mol, smiles)

        # Add hydrogens
        mol.AddHydrogens()

        # Get coordinates for a 3-D structure
        builder = openbabel.OBBuilder()
        builder.Build(mol)

        molfile = obConversion.WriteString(mol)
        self.from_molfile_text(molfile)
        if name is not None:
            self.name = name



# -*- coding: utf-8 -*-

"""Functions for handling MDL molfiles"""

import logging
import time

logger = logging.getLogger(__name__)


class MolFileMixin:
    """A mixin for handling MDL Molfiles."""

    def to_molfile_text(
        self, configuration=None, title=None, comment='Exported from SEAMM'
    ):
        """Create the text of the Molfile from the system.

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.
        title : str = None
            The title for the structure, by default the system name.
        comment : str = 'Exported from SEAMM'
            Comment line

        Returns
        -------
        text : str
            The text of the file.
        """

        lines = []

        atoms = self.atoms
        bonds = self.bonds

        if configuration is None:
            configuration = self.current_configuration
        n_atoms = atoms.n_atoms(configuration=configuration)
        n_bonds = bonds.n_bonds(configuration=configuration)

        nsgroups = 0
        n3d = 0
        is_chiral = 0  # may need to think about this later.

        if title is None:
            lines.append(self.name)
        else:
            lines.append(title)
        date_time = time.strftime('%m%d%y%H%M')

        lines.append('PS' + 'SEAMM_WF' + date_time + '3D')
        lines.append(comment)
        lines.append('  0  0  0     0  0            999 V3000')

        lines.append('M  V30 BEGIN CTAB')
        lines.append(
            'M  V30 COUNTS {} {} {} {} {}'.format(
                n_atoms, n_bonds, nsgroups, n3d, is_chiral
            )
        )
        lines.append('M  V30 BEGIN ATOM')
        count = 0
        if 'formal charges' in atoms:
            for row in atoms.atoms(configuration):
                count += 1
                symbol = self.to_symbols([row['atno']])
                lines.append(
                    f"M  V30 {count} {symbol} {row['x']} {row['y']} {row['z']}"
                    " 0 CHG={row['formal charge']}"
                )
        else:
            for row in atoms.atoms(configuration=configuration):
                count += 1
                symbol = self.to_symbols([row['atno']])[0]
                lines.append(
                    f"M  V30 {count} {symbol} {row['x']} {row['y']} {row['z']}"
                    " 0"
                )
        lines.append('M  V30 END ATOM')
        lines.append('M  V30 BEGIN BOND')
        count = 0
        for row in bonds.bonds(configuration=configuration):
            count += 1
            lines.append(
                f"M  V30 {count} {row['bondorder']} "
                f"{row['i']} {row['j']}"
            )
        lines.append('M  V30 END BOND')
        lines.append('M  V30 END CTAB')
        lines.append('M  END')

        return '\n'.join(lines)

    def from_molfile_text(self, data, configuration=None):
        """Create the system from an MDL Molfile, version 3

        Parameters
        ----------
        data : str
            The complete text of the Molfile.
        configuration : int = None
            The configuration to use, defaults to the current configuration.
        """

        self.periodicity = 0

        n_molecules = 0
        lines = enumerate(data.splitlines())

        # title
        lineno, title = next(lines)
        self.name = title.strip()
        # header
        next(lines)
        # comment
        next(lines)
        lineno, line = next(lines)
        if line.split()[6] != 'V3000':
            raise RuntimeError(
                f"molfile:to_seamm -- the file is not version 3: '{line}'"
            )
        for lineno, line in lines:
            logger.debug(f'{lineno}: {line}')
            if 'M  END' in line:
                break
            elif 'M  V30 BEGIN CTAB' in line:
                n_molecules += 1
                if n_molecules > 1:
                    raise NotImplementedError('Multiple molecules?')
            elif 'M V30 END CTAB' in line:
                pass
            elif 'M  V30 COUNTS' in line:
                natoms, nbonds, nsgroups, n3d, is_chiral = line.split()[3:]

                natoms = int(natoms)
                nbonds = int(nbonds)
                # not used, yet.
                # nsgroups = int(nsgroups)
                # n3d = int(n3d)
                # is_chiral = bool(is_chiral)
            elif 'M  V30 BEGIN ATOM' in line:
                logger.debug('In atom table')
                xs = []
                ys = []
                zs = []
                symbols = []
                formal_charges = []
                have_formal_charges = False
                for lineno, line in lines:
                    if 'M  V30 END ATOM' in line:
                        logger.debug(f'Saving {len(xs)} atoms to system')
                        if (
                            have_formal_charges and
                            'formal_charges' not in self.atoms
                        ):
                            logger.debug('   with formal charges')
                            self.atoms.add_attribute(
                                'formal_charge', coltype='int', default=0
                            )
                            atom_ids = self.atoms.append(
                                configuration=configuration,
                                x=xs,
                                y=ys,
                                z=zs,
                                symbol=symbols,
                                formal_charge=formal_charges
                            )
                        else:
                            atom_ids = self.atoms.append(
                                configuration=configuration,
                                x=xs,
                                y=ys,
                                z=zs,
                                symbol=symbols
                            )
                        break
                    i, symbol, x, y, z, q = line.split()[2:8]
                    xs.append(float(x))
                    ys.append(float(y))
                    zs.append(float(z))
                    symbols.append(symbol)
                    if 'CHG=' in line:
                        for tmp in line.split()[8:]:
                            if 'CHG=' in tmp:
                                formal_charges.append(int(tmp[4:]))
                                have_formal_charges = True
                    else:
                        formal_charges.append(0)
            elif 'M  V30 BEGIN BOND' in line:
                logger.debug('In bond table')
                iatoms = []
                jatoms = []
                bondorders = []
                for lineno, line in lines:
                    if 'M  V30 END BOND' in line:
                        if len(iatoms) > 0:
                            self.bonds.append(
                                i=iatoms,
                                j=jatoms,
                                bondorder=bondorders,
                                configuration=configuration
                            )
                        break
                    bondorder, iatom, jatom = line.split()[3:6]
                    iatoms.append(atom_ids[int(iatom) - 1])
                    jatoms.append(atom_ids[int(jatom) - 1])
                    bondorders.append(int(bondorder))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy
import pytest  # noqa: F401
import os
import os.path
import shutil
import tempfile
import time

from molsystem import Systems  # noqa: F401
"""Tests for the System classes."""

natoms = 1000000


@pytest.fixture(scope="module")
def msystem():
    systems = Systems()

    newpath = tempfile.mkdtemp()
    filepath = os.path.join(newpath, 'seamm.db')

    system = systems.create_system('seamm', filename=filepath)

    yield system

    shutil.rmtree(newpath)


@pytest.fixture(scope="module")
def matoms(msystem):
    """An empty atoms table."""
    return msystem['atoms']


@pytest.mark.timing
def test_creation(matoms):
    """Create a system with a million atoms"""
    print(f'\nCreating a system with {natoms} atoms')
    t0 = time.perf_counter()
    rng = numpy.random.default_rng()
    atno = rng.integers(1, 101, size=natoms)
    x = rng.uniform(low=0, high=100, size=natoms)
    y = rng.uniform(low=0, high=100, size=natoms)
    z = rng.uniform(low=0, high=100, size=natoms)
    t1 = time.perf_counter()
    print(f'      setup took {t1-t0:.3} s')

    t0 = time.perf_counter()
    matoms.append(atno=atno.tolist(), x=x.tolist(), y=y.tolist(), z=z.tolist())
    t1 = time.perf_counter()

    print(f'  appending took {t1-t0:.3} s')
    filename = matoms.system.filename
    size = os.path.getsize(filename) / 1024 / 1024
    print(f'       database is {size:.3} MB')

    assert matoms.n_atoms() == natoms


@pytest.mark.timing
def test_access(matoms):
    """Check that we can access the created system"""
    assert matoms.n_atoms() == natoms


@pytest.mark.timing
def test_loop_over_atoms(matoms):
    """Loop over all the atoms (rows)."""
    n = 0
    for atom in matoms.atoms():
        n += 1
    assert n == matoms.n_atoms()


@pytest.mark.timing
def test_selection(matoms):
    """Test selecting a subset of atoms."""
    n = 0
    for atom in matoms.atoms('atno', '==', 6):
        n += 1
    print(f'\nFound {n} carbon atoms in {natoms} atoms')
    print(f'There should be about {int(natoms / 100)}')
    assert atom['atno'] == 6



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding systems
"""

import collections.abc
import logging
from pathlib import Path
import shutil
import sqlite3
import tempfile

import pathvalidate

from molsystem.system import _System

logger = logging.getLogger(__name__)


class Systems(collections.abc.MutableMapping):

    def __init__(self):
        """Initialize the systems with no systems."""

        self._systems = {}

    def __getitem__(self, key):
        """Allow [] access to the dictionary of systems"""
        return self._systems[key]['system']

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        raise NotImplementedError(f"Table '{key}' cannot be created yet")

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self:
            data = self._systems[key]
            del data['system']
            if data['temporary']:
                if 'tempdir' in data:
                    shutil.rmtree(data['tempdir'])
                else:
                    data['path'].unlink()
            del self._systems[key]
        else:
            raise KeyError(
                f"Trying to delete system '{key}', which does not exist."
            )

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._systems)

    def __len__(self):
        """The len() command"""
        return len(self._systems)

    def __repr__(self):
        """The string representation of this object"""
        return repr(self._systems)

    def __str__(self):
        """The pretty string representation of this object"""
        return str(self._systems)

    def __contains__(self, key):
        """Return a boolean indicating if a system exists."""
        return key in self._systems

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        raise NotImplementedError()

    def list(self):
        """Return a list of the systems."""
        return [*self._systems]

    def create_system(self, name, filename=None, temporary=False, force=False):
        """Create a system with a given name, and optionally a filename."""
        if name in self:
            raise KeyError(f"System '{name}' already exists.")

        self._systems[name] = data = {}
        data['temporary'] = temporary

        if temporary:
            tmp = pathvalidate.sanitize_filename(name + '.db', platform='auto')
            data['tempdir'] = Path(tempfile.mkdtemp())
            path = data['tempdir'] / tmp
        elif filename is None:
            tmp = pathvalidate.sanitize_filename(name + '.db', platform='auto')
            path = Path(tmp)
        else:
            tmp = pathvalidate.sanitize_filename(filename, platform='auto')
            path = Path(tmp)

        path = path.expanduser().resolve()
        if path.exists():
            if force:
                path.unlink()
            else:
                raise RuntimeError(f"File '{path}' exists!")

        filename = str(path)
        system = _System(self, nickname=name, filename=filename)

        data['system'] = system
        data['path'] = path

        return system

    def copy_system(
        self, other, name=None, filename=None, temporary=False, force=False
    ):
        """Create a copy of a system, optionally with a given name and
        filename."""

        if name is None:
            tmp_name = other.name
            name = tmp_name + '_copy'
            i = 1
            while name in self:
                i += 1
                name = f'{tmp_name}_copy_{i}'
        elif name in self:
            raise KeyError(f"System '{name}' already exists.")

        self._systems[name] = data = {}
        data['temporary'] = temporary

        if temporary:
            tmp = pathvalidate.sanitize_filename(name + '.db', platform='auto')
            data['tempdir'] = Path(tempfile.mkdtemp())
            path = data['tempdir'] / tmp
        elif filename is None:
            tmp = pathvalidate.sanitize_filename(name + '.db', platform='auto')
            path = Path(tmp)

        # Get simple absolute path
        path = path.expanduser().resolve()

        if path.exists():
            if force:
                path.unlink()
            else:
                raise RuntimeError(f"File '{path}' exists!")

        filename = str(path)

        # Copy the database over
        db = sqlite3.connect(filename)
        other.db.commit()
        other.db.backup(db)
        db.close()

        # and open it
        system = _System(self, nickname=name, filename=filename)

        data['system'] = system
        data['path'] = path

        return system

    def open_system(self, filename, name=None, temporary=False):
        """Open an existing system with a given name."""
        path = Path(filename)
        if name is None:
            name = path.with_suffix('').name

        if name in self:
            raise RuntimeError(f"System '{name}' already exists!")

        path = path.expanduser().resolve()
        if not path.exists():
            raise RuntimeError(f"File '{path}' does not exist!")

        self._systems[name] = data = {}
        data['temporary'] = temporary

        filename = str(path)

        # and open it
        system = _System(self, nickname=name, filename=filename)

        data['system'] = system
        data['path'] = path

        return system

    def overwrite(self, system, other):
        """Overwrite a system with the contents of another."""

        system.db.commit()
        system.cursor.close()
        system.db.close()
        path = self._systems[system.nickname]['path']
        path.unlink()
        system._db = sqlite3.connect(path)
        system._db.row_factory = sqlite3.Row
        system._db.execute('PRAGMA foreign_keys = ON')
        system._cursor = system._db.cursor()
        other.db.commit()
        other.db.backup(system._db)



# -*- coding: utf-8 -*-

"""Topological methods for the system"""

import logging

from openbabel import openbabel

logger = logging.getLogger(__name__)


class TopologyMixin:
    """A mixin for handling topology."""

    def find_molecules(self, configuration=None, as_indices=False):
        """Find the separate molecules in a system.

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.
        as_indices : bool = False
            Whether to return 0-based indices (True) or atom ids (False)

        Returns
        -------
        molecules : [[int]*n_molecules]
            A list of lists of atom ids or indices for the molecules
        """

        if configuration is None:
            configuration = self.current_configuration

        molecules = []

        atoms = self['atoms']
        atom_ids = atoms.atom_ids(configuration)
        n_atoms = len(atom_ids)

        if n_atoms == 0:
            return molecules

        to_index = {j: i for i, j in enumerate(atom_ids)}
        neighbors = self.bonded_neighbors(configuration)
        visited = [False] * n_atoms
        while True:
            # Find first atom not yet visited
            try:
                index = visited.index(False)
            except ValueError:
                break
            visited[index] = True
            i = atom_ids[index]
            atoms = [i]
            next_atoms = neighbors[i]
            while len(next_atoms) > 0:
                tmp = []
                for i in next_atoms:
                    if not visited[to_index[i]]:
                        atoms.append(i)
                        visited[to_index[i]] = True
                        tmp.extend(neighbors[i])
                next_atoms = tmp
            molecules.append(sorted(atoms))
        if as_indices:
            return [[to_index[j] for j in js] for js in molecules]
        else:
            return molecules

    def bonded_neighbors(
        self, configuration=None, as_indices=False, first_index=0
    ):
        """The atoms bonded to each atom in the system.

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.
        as_indices : bool = False
            Whether to return 0-based indices (True) or atom ids (False)
        first_index : int = 0
            The smallest index, e.g. 0 or 1

        Returns
        -------
        neighbors : {int: [int]} or [[int]] for indices
            list of atom ids for each atom id
        """
        neighbors = {}

        atoms = self['atoms']
        bonds = self['bonds']
        n_atoms = atoms.n_atoms(configuration)

        if n_atoms == 0:
            if as_indices:
                return []
            else:
                return neighbors

        atom_ids = atoms.atom_ids(configuration)
        neighbors = {i: [] for i in atom_ids}

        if bonds.n_bonds(configuration) > 0:
            for bond in bonds.bonds(configuration):
                i = bond['i']
                j = bond['j']
                neighbors[i].append(j)
                neighbors[j].append(i)

        if as_indices:
            # Convert to indices
            to_index = {j: i + first_index for i, j in enumerate(atom_ids)}
            result = [[]] * (n_atoms + first_index)
            for i, js in neighbors.items():
                result[to_index[i]] = sorted([to_index[j] for j in js])
            return result
        else:
            for i in neighbors:
                neighbors[i].sort()

            return neighbors

    def create_molecule_subsets(self, configuration=None):
        """Create a subset for each molecule in a configuration.

        By default they all reference an empty template 'all' of type
        'molecule'.

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.

        Returns
        -------
        [int]
            The ids of the subsets, one per molecule.
        """

        if configuration is None:
            configuration = self.current_configuration

        # get the 'all/molecule' template
        templates = self['template']
        tid = templates.find('all', 'molecule', create=True)

        # Find the molecules and the create the subsets if they don't exist.
        molecules = self.find_molecules(configuration=configuration)

        # Remove any previous subsets for this configuration
        subsets = self['subset']
        sids = subsets.find(tid, configuration=configuration)
        if len(sids) > 0:
            subsets.delete(sids)

        # Now create the new set.
        sids = []
        for atom_ids in molecules:
            sid = subsets.create(
                tid, configuration=configuration, atoms=atom_ids
            )
            sids.append(sid)

        return sids

    def create_molecule_templates(
        self, configuration=None, create_subsets=True
    ):
        """Create a template for each unique molecule in a configuration.

        By default also create subsets linking each template to the atoms
        of the molecules in the system.

        Parameters
        ----------
        configuration : int = None
            The configuration to use, defaults to the current configuration.
        create_subsets : bool = True
            If true, create subsets linking the templates to the molecules.

        Returns
        -------
        [int] or [[int], [int]]
            The ids of the templates, or if create_subsets is True
            a two-element list containing the list of templates and
            list of subsets.
        """
        if configuration is None:
            configuration = self.current_configuration

        # Find the molecules
        molecules = self.find_molecules(configuration=configuration)
        n_molecules = len(molecules)

        # And the molecule each atom is in
        atom_to_molecule = {}
        for molecule, atoms in enumerate(molecules):
            for atom in atoms:
                atom_to_molecule[atom] = molecule

        # The bonds in each molecule
        bonds_per_molecule = [[] for i in range(n_molecules)]
        for bond in self.bonds.bonds(configuration=configuration):
            i = bond['i']
            j = bond['j']
            order = bond['bondorder']
            molecule = atom_to_molecule[i]
            bonds_per_molecule[molecule].append((i, j, order))

        # Get the canonical smiles for each molecule
        to_can = openbabel.OBConversion()
        to_can.SetOutFormat('can')
        to_smi = openbabel.OBConversion()
        to_smi.SetOutFormat('smi')
        ob_mol = openbabel.OBMol()
        ob_template = openbabel.OBMol()
        atnos = self.atoms.atomic_numbers(configuration)
        start = 0
        sids = {}
        tids = []
        for molecule, atoms in enumerate(molecules):
            to_index = {j: i for i, j in enumerate(atoms)}
            n_atoms = len(atoms)
            molecule_atnos = atnos[start:start + n_atoms]
            start += n_atoms

            ob_mol.Clear()
            for atom, atno in zip(atoms, molecule_atnos):
                ob_atom = ob_mol.NewAtom()
                ob_atom.SetAtomicNum(atno)
            bonds = []
            for i, j, order in bonds_per_molecule[molecule]:
                bonds.append((to_index[i], to_index[j], order))
                # 1-based indices in ob.
                ob_mol.AddBond(to_index[i] + 1, to_index[j] + 1, order)

            smiles = to_smi.WriteString(ob_mol).strip()
            canonical = to_can.WriteString(ob_mol).strip()

            # See if a molecule template with the canonical smiles exists
            if self.templates.exists(canonical, 'molecule'):
                tid = self.templates.find(canonical, 'molecule')
            else:
                tid = self.templates.create(
                    canonical, 'molecule', atnos=molecule_atnos, bonds=bonds
                )
            tids.append(tid)
            if create_subsets:
                tatom_ids = self.templateatoms.atom_ids(tid)
                if smiles != canonical:
                    # Need to reorder the atoms to match the template atoms

                    # Prepare the OB molecule for the template
                    ob_template.Clear()
                    for atno in self.templateatoms.atomic_numbers(tid):
                        ob_atom = ob_template.NewAtom()
                        ob_atom.SetAtomicNum(atno)

                    tatom_ids = self.templateatoms.atom_ids(tid)
                    to_index = {j: i for i, j in enumerate(tatom_ids)}
                    for row in self.templatebonds.bonds(tid):
                        i = to_index[row['i']]
                        j = to_index[row['j']]
                        order = row['bondorder']
                        ob_template.AddBond(i + 1, j + 1, order)

                    # Get the mapping from template to molecule
                    query = openbabel.CompileMoleculeQuery(ob_template)
                    mapper = openbabel.OBIsomorphismMapper.GetInstance(query)
                    mapping = openbabel.vpairUIntUInt()
                    mapper.MapFirst(ob_mol, mapping)
                    tmp = [atoms[j] for i, j in mapping]
                    atoms = tmp
                sid = self.subsets.create(tid, configuration, atoms, tatom_ids)
                if tid not in sids:
                    sids[tid] = []
                sids[tid].append(sid)

        if create_subsets:
            return tids, sids
        else:
            return tids



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401
"""Tests for handling SMILES."""


def test_to_smiles(AceticAcid):
    """Create a SMILES string from a system"""
    correct = 'CC(=O)O'
    smiles = AceticAcid.to_smiles()

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_to_smiles_with_name(AceticAcid):
    """Create a SMILES string from a system"""
    correct = ['CC(=O)O', 'acetic acid']
    smiles = AceticAcid.to_smiles(name=True)

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_to_canonical_smiles(AceticAcid):
    """Create a SMILES string from a system"""
    correct = 'CC(=O)O'
    smiles = AceticAcid.to_smiles(canonical=True)

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_several_molecules(CH3COOH_3H2O):
    """System with acetic acid and 3 waters"""
    system = CH3COOH_3H2O
    correct = 'CC(=O)O.O.O.O'

    smiles = system.to_smiles()

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_several_molecules_canonical(CH3COOH_3H2O):
    """System with acetic acid and 3 waters"""
    system = CH3COOH_3H2O
    correct = 'CC(=O)O.O.O.O'

    smiles = system.to_smiles(canonical=True)

    if smiles != correct:
        print(smiles)
    assert smiles == correct


def test_from_smiles(system):
    """Create a system from SMILES"""
    correct = ['CC(=O)O', 'acetic acid']
    system.from_smiles('OC(=O)C', name='acetic acid')
    result = system.to_smiles(name=True, canonical=True)

    if result != correct:
        print(result)

    assert result == correct



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401
"""Tests for handling Molfiles."""


def test_to_mmcif_text(AceticAcid):
    """Write a manually created system to mmcif"""
    correct = """\
# Generated by MolSSI SEAMM
data_CH2O
_chem_comp.name 'C2 H4 O2'
_chem_comp.id 'CH2O'
_chem_comp.formula   'C2 H4 O2'
_chemical_formula_structural   CH2O
_chemical_formula_sum   'C2 H4 O2'
loop_
 _atom_site_type_symbol
 _atom_site_label
 _atom_site_symmetry_multiplicity
 _atom_site_fract_x
 _atom_site_fract_y
 _atom_site_fract_z
 _atom_site_occupancy
MOL1 C C 1.080 0.018 -0.018 1.080 0.018 -0.018 HET 1
MOL1 H H 0.578 3.138 0.281 0.578 3.138 0.281 HET 1
MOL1 H2 H 0.721 -0.674 -0.786 0.721 -0.674 -0.786 HET 1
MOL1 H3 H 0.705 -0.314 0.953 0.705 -0.314 0.953 HET 1
MOL1 C2 C 0.571 1.390 -0.316 0.571 1.390 -0.316 HET 1
MOL1 O O -0.132 1.714 -1.257 -0.132 1.714 -1.257 HET 1
MOL1 O2 O 0.976 2.297 0.592 0.976 2.297 0.592 HET 1
MOL1 H4 H 2.172 0.016 -0.031 2.172 0.016 -0.031 HET 1
#
loop_
 _chem_comp_bond.comp_id
 _chem_comp_bond.atom_id_1
 _chem_comp_bond.atom_id_2
 _chem_comp_bond.value_order
MOL1 C H SING
MOL1 C H2 SING
MOL1 C H3 SING
MOL1 C C2 SING
MOL1 C2 O DOUB
MOL1 C2 O2 SING
MOL1 O2 H4 SING"""

    text = AceticAcid.to_mmcif_text()

    if text != correct:
        print(text)

    assert text == correct


def test_to_cif_text(vanadium):
    """Write a manually created system to mmcif"""
    correct = """\
# Generated by MolSSI SEAMM
data_V
_symmetry_space_group_name_H-M   'P 1'
_cell_length_a   3.03
_cell_length_b   3.03
_cell_length_c   3.03
_cell_angle_alpha   90.0
_cell_angle_beta    90.0
_cell_angle_gamma   90.0
_symmetry_Int_Tables_number   1
_cell_volume   27.818126999999997
_cell_formula_units_Z   2
loop_
 _symmetry_equiv_pos_site_id
 _symmetry_equiv_pos_as_xyz
  1  'x, y, z'
_chemical_formula_structural   V
_chemical_formula_sum   'V2'
loop_
 _atom_site_type_symbol
 _atom_site_label
 _atom_site_symmetry_multiplicity
 _atom_site_fract_x
 _atom_site_fract_y
 _atom_site_fract_z
 _atom_site_occupancy
V V  1  0.000 0.000 0.000  1
V V2  1  0.500 0.500 0.500  1"""

    text = vanadium.to_cif_text()

    if text != correct:
        print(text)

    assert text == correct



# -*- coding: utf-8 -*-

"""
molsystem
A general implementation for molecular and periodic systems.
"""

# Bring up the classes so that they appear to be directly in
# the molsystem package.

from molsystem.systems import Systems  # noqa: F401
from molsystem.cell import Cell  # noqa: F401

# Handle versioneer
from ._version import get_versions
__author__ = """Paul Saxe"""
__email__ = 'psaxe@molssi.org'
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions
